"""A module to contruct layers in a DAG for inspiral workflows."""

# Copyright (C) 2020       Patrick Godwin
# Copyright (C) 2024-2025  Yun-Jing Huang, Cort Posnansky

import itertools
import os
import shutil
from collections.abc import Mapping
from math import ceil
from typing import Iterable

import numpy
from ezdag import Argument, Layer, Node, Option
from lal import rate

from sgnl.dags import util
from sgnl.dags.util import condor_scratch_space, format_ifo_args, groups, to_ifo_list

DEFAULT_MAX_FILES = 500


def create_layer(executable, condor_config, resource_requests, retries=3, name=None):
    # Default submit file options
    submit_description = create_submit_description(condor_config)

    # Add resource requests
    submit_description.update(resource_requests)

    # Allow arbitrary submit file options from the config
    if executable in condor_config:
        submit_description.update(condor_config[executable])

    # Set file transfer
    if condor_config.transfer_files is not None:
        transfer_files = condor_config.transfer_files
    else:
        transfer_files = True

    return Layer(
        executable,
        name=name if name else executable,
        transfer_files=transfer_files,
        submit_description=submit_description,
        retries=retries,
    )


def create_submit_description(condor_config):
    submit_description = {
        "want_graceful_removal": "True",
        "kill_sig": "15",
        "accounting_group": condor_config.accounting_group,
        "accounting_group_user": condor_config.accounting_group_user,
    }

    if condor_config.getenv:
        submit_description["getenv"] = condor_config.getenv

    requirements = []
    environment = {}

    # Container options
    if "container" in condor_config:
        submit_description["MY.SingularityImage"] = f'"{condor_config.container}"'
        submit_description["transfer_executable"] = False

    # Scitoken options
    if condor_config.use_scitokens is True:
        submit_description["use_oauth_services"] = "scitokens"
        environment["BEARER_TOKEN_FILE"] = (
            f"$$(CondorScratchDir)/.condor_creds/"
            f"{submit_description['use_oauth_services']}.use"
        )

    # Config options
    if "directives" in condor_config:
        submit_description.update(condor_config["directives"])
    if "requirements" in condor_config:
        requirements.extend(condor_config["requirements"])
    if "environment" in condor_config:
        environment.update(condor_config["environment"])

    # Condor requirements
    submit_description["requirements"] = " && ".join(requirements)

    # Condor environment
    env_opts = [f"{key}={val}" for (key, val) in environment.items()]
    if "environment" in submit_description:
        env_opts.append(submit_description["environment"].strip('"'))
    submit_description["environment"] = f'"{" ".join(env_opts)}"'

    return submit_description


def test(echo_config, condor_config):
    # FIXME Delete this layer
    executable = "echo"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "10MB",
        "request_disk": "10MB",
    }
    layer = create_layer(executable, condor_config, resource_requests)

    for i in range(echo_config.jobs):
        arguments = [Argument("words", f"This is job {i}")]
        inputs = [Argument("file_in", "inputs/file.txt")]
        outputs = [Argument("file_out", "outputs/output.txt")]
        layer += Node(arguments=arguments, inputs=inputs, outputs=outputs)

    return layer


def reference_psd(
    filter_config, psd_config, source_config, condor_config, ref_psd_cache
):
    executable = "sgnl-reference-psd"
    resource_requests = {
        "request_cpus": 2,
        "request_memory": "2GB",
        "request_disk": "2GB",
    }
    layer = create_layer(executable, condor_config, resource_requests)

    common_opts = [
        Option("data-source", "frames"),
        Option("psd-fft-length", psd_config.fft_length),
        Option("frame-segments-name", source_config.frame_segments_name),
    ]

    if filter_config.search:
        common_opts.append(Option("search", filter_config.search))

    for (ifo_combo, span), psds in ref_psd_cache.groupby("ifo", "time").items():
        ifos = to_ifo_list(ifo_combo)
        start, end = span

        arguments = [
            Option("gps-start-time", int(start)),
            Option("gps-end-time", int(end)),
            Option("channel-name", format_ifo_args(ifos, source_config.channel_name)),
            *common_opts,
        ]
        inputs = [
            Option(
                "frame-segments-file", source_config.frame_segments_file, track=False
            )
        ]

        if source_config.frame_cache:
            inputs.append(Option("frame-cache", source_config.frame_cache, track=False))
        else:
            arguments.extend(
                [
                    Option(
                        "frame-type", format_ifo_args(ifos, source_config.frame_type)
                    ),
                    Option("data-find-server", source_config.data_find_server),
                ]
            )

        layer += Node(
            arguments=arguments,
            inputs=inputs,
            outputs=Option("output-name", psds.files),
        )

    return layer


def median_psd(psd_config, condor_config, ref_psd_cache, median_psd_cache):
    executable = "sgnl-median-of-psds"
    resource_requests = {
        "request_cpus": 2,
        "request_memory": "2GB",
        "request_disk": "2GB",
    }
    layer = create_layer(executable, condor_config, resource_requests)

    layer += Node(
        inputs=Option("input-files", ref_psd_cache.files),
        outputs=Option("output-name", median_psd_cache.files),
    )

    return layer


def svd_bank(
    svd_config,
    condor_config,
    all_ifos,
    split_bank_cache,
    median_psd_cache,
    svd_cache,
    svd_bins,
    svd_stats,
):
    executable = "sgnl-inspiral-svd-bank"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "7GB",
        "request_disk": "2GB",
    }
    layer = create_layer(executable, condor_config, resource_requests)

    arguments = [
        Option("f-low", svd_config.f_low),
        Option("f-final", svd_config.max_f_final),
        Option("approximant", svd_config.approximant),
        Option("overlap", svd_config.overlap),
        Option("instrument", to_ifo_list(all_ifos)),
        Option("n", svd_config.num_split_templates),
        Option("num-banks", svd_config.num_banks),
        Option("sort-by", svd_config.sort_by),
    ]

    if svd_config.autocorrelation_length:
        mchirp_to_ac_length = autocorrelation_length_map(
            svd_config.autocorrelation_length
        )

    split_banks = split_bank_cache.groupby("bin")
    for (ifo, svd_bin), svd_banks in svd_cache.groupby("ifo", "bin").items():

        # grab sub-bank specific configuration if available
        if "bank_name" in svd_stats["bins"][svd_bin]:
            bank_name = svd_stats["bins"][svd_bin]["bank_name"]
            this_svd_config = svd_config.sub_banks[bank_name]
        else:
            this_svd_config = svd_config

        arguments = [
            Option("instrument-override", ifo),
            Option("flow", this_svd_config.f_low),
            Option("samples-min", this_svd_config.samples_min),
            Option("samples-max-64", this_svd_config.samples_max_64),
            Option("samples-max-256", this_svd_config.samples_max_256),
            Option("samples-max", this_svd_config.samples_max),
            Option("svd-tolerance", this_svd_config.tolerance),
        ]

        # FIXME: in gstlal offline, autocrrelation_length is in the option file,
        # but in online, it's in the config. Which one should we use?
        if this_svd_config.autocorrelation_length:
            bin_mchirp = svd_stats["bins"][svd_bin]["mean_mchirp"]
            arguments.append(
                Option("autocorrelation-length", mchirp_to_ac_length(bin_mchirp))
            )
            if "ac_length" in svd_stats["bins"][svd_bin]:
                # FIXME: sanity check autocorrelation length is the same, remove
                # this once we decide on a single location to get the
                # autocorrelation length
                assert svd_stats["bins"][svd_bin]["ac_length"] == mchirp_to_ac_length(
                    bin_mchirp
                )
        elif "ac_length" in svd_stats["bins"][svd_bin]:
            arguments.append(
                Option(
                    "autocorrelation-length", svd_stats["bins"][svd_bin]["ac_length"]
                )
            )
        else:
            raise ValueError("Unknown autocorrelation length option")
        if "max_duration" in this_svd_config:
            arguments.append(Option("max-duration", this_svd_config.max_duration))
        if "sample_rate" in this_svd_config:
            arguments.append(Option("sample-rate", this_svd_config.sample_rate))
        # FIXME figure out where this option should live
        # if 'use_bankchisq' in config.rank and config.rank.use_bankchisq:
        #    arguments.append(Option("use-bankchisq"))

        layer += Node(
            arguments=arguments,
            inputs=[
                Option("reference-psd", median_psd_cache.files),
                Option("template-banks", sorted(split_banks[svd_bin].files)),
            ],
            outputs=Option("write-svd", svd_banks.files),
        )

    return layer


def filter(
    psd_config,
    svd_config,
    filter_config,
    source_config,
    condor_config,
    ref_psd_cache,
    svd_bank_cache,
    lr_cache,
    trigger_cache,
    svd_stats,
    min_instruments,
):
    executable = "sgnl-inspiral"
    resource_requests = {
        "request_cpus": 2,
        "request_memory": "4GB",
        "request_disk": "5GB",
    }

    # Set torch-device
    if filter_config.torch_device:
        torch_device = filter_config.torch_device
    else:
        torch_device = "cpu"

    if "cuda" in torch_device:
        resource_requests["request_gpus"] = 1
        resource_requests["request_cpus"] = 1

    layer = create_layer(executable, condor_config, resource_requests)

    common_opts = [
        Option("track-psd"),
        Option("data-source", "frames"),
        Option("psd-fft-length", psd_config.fft_length),
        Option("frame-segments-name", source_config.frame_segments_name),
        Option("coincidence-threshold", filter_config.coincidence_threshold),
        Option("snr-min", filter_config.snr_min),
    ]

    common_opts.append(Option("min-instruments-candidates", min_instruments))

    if filter_config.all_triggers_to_background:
        common_opts.append(Option("all-triggers-to-background"))

    if filter_config.search:
        common_opts.append(Option("search", filter_config.search))

    common_inputs = [
        Option("event-config", filter_config.event_config_file),
    ]

    # Set torch-dtype
    if filter_config.torch_dtype:
        torch_dtype = filter_config.torch_dtype
    else:
        torch_dtype = "float32"
    common_opts.append(Option("torch-dtype", torch_dtype))

    common_opts.append(Option("torch-device", torch_device))

    # Add gstlal CPU upsampling option if enabled
    if filter_config.use_gstlal_cpu_upsample:
        common_opts.append(Option("use-gstlal-cpu-upsample"))

    # Set trigger-finding-duration
    if filter_config.trigger_finding_duration:
        trigger_finding_duration = filter_config.trigger_finding_duration
    else:
        trigger_finding_duration = 1
    common_opts.append(Option("trigger-finding-duration", trigger_finding_duration))

    # Checkpoint by grouping SVD bins together
    if filter_config.group_svd_num:
        group_svd_num = filter_config.group_svd_num
    else:
        group_svd_num = 1

    ref_psds = ref_psd_cache.groupby("ifo", "time")
    svd_banks = svd_bank_cache.groupby("ifo", "bin")
    lrs = lr_cache.groupby("ifo", "time", "bin")
    for (ifo_combo, span), triggers in trigger_cache.groupby("ifo", "time").items():
        ifos = to_ifo_list(ifo_combo)
        start, end = span

        filter_opts = [
            Option("gps-start-time", int(start)),
            Option("gps-end-time", int(end)),
            Option("channel-name", format_ifo_args(ifos, source_config.channel_name)),
        ]
        inputs = [
            Option("frame-segments-file", source_config.frame_segments_file),
            # Option("veto-segments-file", filter_config.veto_segments_file),
            Option("reference-psd", ref_psds[(ifo_combo, span)].files),
            # Option("time-slide-file", filter_config.time_slide_file),
        ]

        if source_config.frame_cache:
            inputs.append(Option("frame-cache", source_config.frame_cache, track=False))
        else:
            filter_opts.extend(
                [
                    Option(
                        "frame-type", format_ifo_args(ifos, source_config.frame_type)
                    ),
                    Option("data-find-server", source_config.data_find_server),
                ]
            )

        if source_config.idq_channel_name:
            filter_opts.append(
                Option(
                    "idq-channel-name",
                    format_ifo_args(ifos, source_config.idq_channel_name),
                )
            )

        filter_opts.extend(common_opts)

        if source_config.idq_channel_name and filter_config.idq_gate_threshold:
            filter_opts.append(
                Option("idq-gate-threshold", filter_config.idq_gate_threshold)
            )

        if source_config.idq_channel_name and source_config.idq_state_channel_name:
            filter_opts.append(
                Option(
                    "idq-state-channel-name",
                    format_ifo_args(ifos, source_config.idq_state_channel_name),
                )
            )

        for trigger_group in triggers.chunked(group_svd_num):
            svd_bins = trigger_group.groupby("bin").keys()

            thresholds = [
                svd_stats.bins[svd_bin]["ht_gate_threshold"] for svd_bin in svd_bins
            ]
            these_opts = [Option("ht-gate-threshold", thresholds), *filter_opts]

            svd_bank_files = util.flatten(
                [
                    svd_banks[(ifo, svd_bin)].files
                    for ifo in ifos
                    for svd_bin in svd_bins
                ]
            )
            lr_files = util.flatten(
                [lrs[(ifo_combo, span, svd_bin)].files for svd_bin in svd_bins]
            )

            layer += Node(
                arguments=these_opts,
                inputs=[
                    Option("svd-bank", svd_bank_files),
                    *inputs,
                    *common_inputs,
                ],
                outputs=[
                    Option("trigger-output", trigger_group.files),
                    Option("output-likelihood-file", lr_files),
                ],
            )

    return layer


def injection_filter(
    psd_config,
    svd_config,
    filter_config,
    injection_config,
    source_config,
    condor_config,
    ref_psd_cache,
    svd_bank_cache,
    trigger_cache,
    svd_stats,
    min_instruments,
):
    executable = "sgnl-inspiral"
    resource_requests = {
        "request_cpus": 2,
        "request_memory": "5GB",
        "request_disk": "5GB",
    }
    # Set torch-device
    if filter_config.torch_device:
        torch_device = filter_config.torch_device
    else:
        torch_device = "cpu"

    if "cuda" in torch_device:
        resource_requests["request_gpus"] = 1
        resource_requests["request_cpus"] = 1

    layer = create_layer(
        executable, condor_config, resource_requests, name="sgnl-inspiral-inj"
    )

    common_opts = [
        Option("track-psd"),
        Option("data-source", "frames"),
        Option("psd-fft-length", psd_config.fft_length),
        Option("frame-segments-name", source_config.frame_segments_name),
        Option("coincidence-threshold", filter_config.coincidence_threshold),
        Option("snr-min", filter_config.snr_min),
        Option("injections"),
    ]

    common_opts.append(Option("min-instruments-candidates", min_instruments))

    common_inputs = [
        Option("event-config", filter_config.event_config_file),
    ]

    if filter_config.search:
        common_opts.append(Option("search", filter_config.search))

    # Set torch-dtype
    if filter_config.torch_dtype:
        torch_dtype = filter_config.torch_dtype
    else:
        torch_dtype = "float32"
    common_opts.append(Option("torch-dtype", torch_dtype))

    common_opts.append(Option("torch-device", torch_device))

    # Add gstlal CPU upsampling option if enabled
    if filter_config.use_gstlal_cpu_upsample:
        common_opts.append(Option("use-gstlal-cpu-upsample"))

    # Set trigger-finding-duration
    if filter_config.trigger_finding_duration:
        trigger_finding_duration = filter_config.trigger_finding_duration
    else:
        trigger_finding_duration = 1
    common_opts.append(Option("trigger-finding-duration", trigger_finding_duration))

    # Checkpoint by grouping SVD bins together
    if filter_config.group_svd_num:
        group_svd_num = filter_config.group_svd_num
    else:
        group_svd_num = 1

    ref_psds = ref_psd_cache.groupby("ifo", "time")
    svd_banks = svd_bank_cache.groupby("ifo", "bin")
    for (ifo_combo, span, inj_type), triggers in trigger_cache.groupby(
        "ifo", "time", "subtype"
    ).items():
        ifos = to_ifo_list(ifo_combo)
        start, end = span
        injection_file = injection_config.filter[inj_type]["file"]

        filter_opts = [
            Option("gps-start-time", int(start)),
            Option("gps-end-time", int(end)),
        ]
        inputs = [
            Option("frame-segments-file", source_config.frame_segments_file),
            Option("reference-psd", ref_psds[(ifo_combo, span)].files),
            Option("injection-file", injection_file),
            # Option("time-slide-file", filter_config.time_slide_file),
        ]

        if (
            source_config.inj_frame_cache
            and injection_config.filter[inj_type]["noiseless_inj_frames"]
        ):
            inputs.append(Option("frame-cache", source_config.frame_cache, track=False))
            inputs.append(
                Option(
                    "noiseless-inj-frame-cache",
                    source_config.inj_frame_cache,
                    track=False,
                )
            )
            filter_opts.append(
                Option(
                    "noiseless-inj-channel-name",
                    format_ifo_args(ifos, source_config.inj_channel_name),
                )
            )
            filter_opts.append(
                Option(
                    "channel-name", format_ifo_args(ifos, source_config.channel_name)
                )
            )
        elif (
            source_config.inj_frame_cache
            and not injection_config.filter[inj_type]["noiseless_inj_frames"]
        ):
            inputs.append(
                Option("frame-cache", source_config.inj_frame_cache, track=False)
            )
            filter_opts.append(
                Option(
                    "channel-name",
                    format_ifo_args(ifos, source_config.inj_channel_name),
                )
            )
        elif source_config.frame_cache and not source_config.inj_frame_cache:
            inputs.append(Option("frame-cache", source_config.frame_cache, track=False))
            filter_opts.append(
                Option(
                    "channel-name", format_ifo_args(ifos, source_config.channel_name)
                )
            )
        else:
            filter_opts.extend(
                [
                    Option(
                        "frame-type", format_ifo_args(ifos, source_config.frame_type)
                    ),
                    Option("data-find-server", source_config.data_find_server),
                    Option(
                        "channel-name",
                        format_ifo_args(ifos, source_config.channel_name),
                    ),
                ]
            )

        if source_config.idq_channel_name:
            filter_opts.append(
                Option(
                    "idq-channel-name",
                    format_ifo_args(ifos, source_config.idq_channel_name),
                )
            )

        filter_opts.extend(common_opts)

        if source_config.idq_channel_name and filter_config.idq_gate_threshold:
            filter_opts.append(
                Option("idq-gate-threshold", filter_config.idq_gate_threshold)
            )

        if source_config.idq_channel_name and source_config.idq_state_channel_name:
            filter_opts.append(
                Option(
                    "idq-state-channel-name",
                    format_ifo_args(ifos, source_config.idq_state_channel_name),
                )
            )

        for trigger_group in triggers.chunked(group_svd_num):
            svd_bins = trigger_group.groupby("bin").keys()

            thresholds = [
                svd_stats.bins[svd_bin]["ht_gate_threshold"] for svd_bin in svd_bins
            ]
            these_opts = [Option("ht-gate-threshold", thresholds), *filter_opts]

            svd_bank_files = util.flatten(
                [
                    svd_banks[(ifo, svd_bin)].files
                    for ifo in ifos
                    for svd_bin in svd_bins
                ]
            )

            layer += Node(
                arguments=these_opts,
                inputs=[
                    Option("svd-bank", svd_bank_files),
                    *inputs,
                    *common_inputs,
                ],
                outputs=[
                    Option("trigger-output", trigger_group.files),
                ],
            )

    return layer


def aggregate(filter_config, condor_config, trigger_cache, clustered_triggers_cache):
    executable = "stillsuit-merge-reduce"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "2GB",
        "request_disk": "1GB",
    }

    layer = create_layer(executable, condor_config, resource_requests)

    clustered_triggers = clustered_triggers_cache.groupby("ifo", "time", "bin")

    # cluster triggers by SNR
    for (ifo_combo, span, svd_bin), triggers in trigger_cache.groupby(
        "ifo", "time", "bin"
    ).items():
        layer += Node(
            arguments=[
                Option("clustering-column", "network_chisq_weighted_snr"),
                Option("clustering-window", 0.1),
            ],
            inputs=[
                Option("config-schema", filter_config.event_config_file),
                Option("dbs", triggers.files),
            ],
            outputs=Option(
                "db-to-insert-into",
                clustered_triggers[(ifo_combo, span, svd_bin)].files,
            ),
        )
    return layer


def marginalize_likelihood_ratio(
    condor_config,
    lr_cache,
    marg_lr_cache,
    prior_cache=None,
):

    executable = "strike-marginalize-likelihood"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "2GB",
        "request_disk": "3GB",
    }
    if prior_cache is not None:
        name = executable + "-prior"
        prior = prior_cache.groupby("bin")
    else:
        name = executable + "-likelihood-ratio-across-time"
    layer = create_layer(
        executable,
        condor_config,
        resource_requests,
        name=name,
    )

    lrs = lr_cache.groupby("bin")

    for svd_bin, marg_lrs in marg_lr_cache.groupby("bin").items():
        layer += Node(
            arguments=Option("marginalize", "likelihood-ratio"),
            inputs=[
                Option(
                    "input",
                    (
                        lrs[svd_bin].files + prior[svd_bin].files
                        if prior_cache is not None
                        else lrs[svd_bin].files
                    ),
                )
            ],
            outputs=Option("output", marg_lrs.files),
        )

    return layer


def create_prior(
    filter_config,
    condor_config,
    prior_config,
    coincidence_threshold,
    svd_bank_cache,
    prior_cache,
    ifos,
    min_instruments,
    svd_stats,
    write_empty_zerolag=None,
    write_empty_marg_zerolag=None,
):
    executable = "sgnl-create-prior-diststats"
    resource_requests = {
        "request_cpus": 2,
        "request_memory": "4GB",
        "request_disk": "3GB",
    }
    layer = create_layer(executable, condor_config, resource_requests)

    svd_banks = svd_bank_cache.groupby("bin")
    arguments = [
        Option("instrument", ifos),
        Option("min-instruments", min_instruments),
        Option("coincidence-threshold", coincidence_threshold),
    ]
    if filter_config.search:
        arguments.append(Option("search", filter_config.search))
    if write_empty_zerolag:
        zerolag_pdf = write_empty_zerolag.groupby("bin")
    if write_empty_marg_zerolag:
        marg_zerolag_pdf = write_empty_marg_zerolag.files[0]

    for i, (svd_bin, prior) in enumerate(prior_cache.groupby("bin").items()):
        inputs = [
            Option("svd-file", svd_banks[svd_bin].files),
            *add_likelihood_ratio_file_options(svd_bin, svd_stats, prior_config),
        ]
        outputs = [Option("output-likelihood-file", prior.files)]
        if write_empty_zerolag and write_empty_marg_zerolag:
            # create an empty rankingstatpdf using the first svd bin's prior file
            # as a bootstrap. This is meant to be used during the setup stage of
            # an online analysis to create the zerolag pdf file
            if i == 0:
                outputs += [
                    Option(
                        "write-empty-rankingstatpdf",
                        zerolag_pdf[(svd_bin)].files + [marg_zerolag_pdf],
                    )
                ]
            else:
                outputs += [
                    Option(
                        "write-empty-rankingstatpdf",
                        zerolag_pdf[(svd_bin)].files + [" /dev/null"],
                    )
                ]

        layer += Node(
            arguments=arguments,
            inputs=inputs,
            outputs=outputs,
        )

    return layer


def add_trigger_dbs(
    condor_config, filter_config, trigger_cache, clustered_trigger_cache, column, window
):
    executable = "stillsuit-merge-reduce"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "6GB",
        "request_disk": "2GB",
    }
    layer = create_layer(
        executable,
        condor_config,
        resource_requests,
        name=executable + "-" + column.replace("_", "-"),
    )

    triggers = trigger_cache.groupby("bin", "subtype")

    # cluster triggers by SNR across time
    for (svd_bin, subtype), clustered_triggers in clustered_trigger_cache.groupby(
        "bin", "subtype"
    ).items():
        layer += Node(
            arguments=[
                Option("clustering-column", column),
                Option("clustering-window", window),
            ],
            inputs=[
                Option("config-schema", filter_config.event_config_file),
                Option("dbs", triggers[(svd_bin, subtype)].files),
            ],
            outputs=Option(
                "db-to-insert-into",
                clustered_triggers.files,
            ),
        )
    return layer


def merge_and_reduce(
    condor_config, filter_config, trigger_cache, clustered_trigger_cache, column, window
):
    executable = "stillsuit-merge-reduce"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "6GB",
        "request_disk": "2GB",
    }
    layer = create_layer(
        executable,
        condor_config,
        resource_requests,
        name=executable + "-" + column.replace("_", "-"),
    )

    triggers = trigger_cache.groupby("subtype")

    # cluster triggers by SNR across time
    for (subtype), clustered_triggers in clustered_trigger_cache.groupby(
        "subtype"
    ).items():
        layer += Node(
            arguments=[
                Option("clustering-column", column),
                Option("clustering-window", window),
            ],
            inputs=[
                Option("config-schema", filter_config.event_config_file),
                Option("dbs", triggers[(subtype)].files),
            ],
            outputs=Option(
                "db-to-insert-into",
                clustered_triggers.files,
            ),
        )
    return layer


def assign_likelihood(
    condor_config,
    filter_config,
    prior_config,
    trigger_cache,
    lr_cache,
    lr_trigger_cache,
    svd_stats,
):
    executable = "sgnl-assign-likelihood"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "6GB",
        "request_disk": "3GB",
    }

    layer = create_layer(executable, condor_config, resource_requests)

    # assign likelihood to triggers
    lr_triggers = lr_trigger_cache.groupby("bin", "subtype")
    lrs = lr_cache.groupby("bin")
    for (svd_bin, subtype), triggers in trigger_cache.groupby("bin", "subtype").items():

        # find path relative to current directory
        # where assigned triggers will reside
        # split_dirname = dirname.split(os.sep)
        # dir_idx = split_dirname.index("triggers")
        # calc_dirname = os.path.join(config.data.rank_dir, *split_dirname[dir_idx:])

        arguments = [
            Option("force"),
            Option("tmp-space", condor_scratch_space()),
        ]

        # allow impossible candidates for inj jobs
        # if config.rank.allow_impossible_inj_candidates:
        #    if inj_type:
        #        arguments.append(Option("allow-impossible-candidates", "True"))
        #    else:
        #        arguments.append(Option("allow-impossible-candidates", "False"))

        layer += Node(
            arguments=arguments,
            inputs=[
                Option("config-schema", filter_config.event_config_file),
                Option("input-database-file", triggers.files),
                Option("input-likelihood-file", lrs[svd_bin].files),
                *add_likelihood_ratio_file_options(
                    svd_bin, svd_stats, prior_config, transfer_only=True
                ),
            ],
            outputs=Option(
                "output-database-file",
                lr_triggers[(svd_bin, subtype)].files,
            ),
        )

    return layer


def calc_pdf(
    condor_config,
    prior_config,
    rank_config,
    config_svd_bins,
    lr_cache,
    pdf_cache,
    svd_stats,
):
    # FIXME: expose this in configuration
    num_cores = rank_config.calc_pdf_cores if rank_config.calc_pdf_cores else 1

    executable = "strike-calc-rank-pdfs"
    resource_requests = {
        "request_cpus": num_cores,
        "request_memory": "3GB",
        "request_disk": "3GB",
    }

    layer = create_layer(executable, condor_config, resource_requests)

    lrs = lr_cache.groupby("bin")
    arguments = [
        Option("num-samples", rank_config.ranking_stat_samples),
        Option("num-cores", num_cores),
    ]
    for svd_bin, pdfs in pdf_cache.groupby("bin").items():
        for pdf in pdfs.files:
            layer += Node(
                arguments=arguments,
                inputs=[
                    Option("input-likelihood-file", lrs[svd_bin].files),
                    *add_likelihood_ratio_file_options(
                        svd_bin, svd_stats, prior_config, transfer_only=True
                    ),
                ],
                outputs=Option("output-rankingstatpdf-file", pdf),
            )

    return layer


def extinct_bin(
    condor_config, event_config_file, pdf_cache, trigger_cache, extinct_cache
):
    executable = "sgnl-extinct-bin"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "4GB",
        "request_disk": "1GB",
    }

    layer = create_layer(executable, condor_config, resource_requests)

    trigger_cache = trigger_cache.groupby("subtype")[""]  # noninj triggers
    trigger_cache = trigger_cache.groupby("bin")
    extinct_pdfs = extinct_cache.groupby("bin")
    for svd_bin, pdfs in pdf_cache.groupby("bin").items():
        assert (
            len(trigger_cache[svd_bin].files) == 1
        ), "Must provide exactly one trigger file per bin to the extinct_bin layer"
        trigger_file = trigger_cache[svd_bin].files[0]

        layer += Node(
            arguments=[Option("reset-zerolag")],
            inputs=[
                Option("config-schema", event_config_file),
                Option("input-database-file", trigger_file),
                Option("input-rankingstatpdf-file", pdfs.files),
            ],
            outputs=Option("output-rankingstatpdf-file", extinct_pdfs[svd_bin].files),
        )

    return layer


def marginalize_pdf(
    condor_config, rank_config, rank_dir, all_ifos, span, pdf_cache, marg_pdf_cache
):
    executable = "strike-marginalize-likelihood"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "2GB",
        "request_disk": "1GB",
    }
    round1_layer = create_layer(
        executable, condor_config, resource_requests, name=executable + "-pdf-round-one"
    )

    num_files = (
        rank_config.marg_pdf_files if rank_config.marg_pdf_files else DEFAULT_MAX_FILES
    )

    # if number of bins is large, combine in two stages instead
    if len(pdf_cache.files) > num_files:
        partial_pdf_files = []
        for pdf_subset in pdf_cache.chunked(num_files):

            # determine bin range and determine partial file name
            svd_bins = list(pdf_subset.groupby("bin").keys())
            min_bin, max_bin = min(svd_bins), max(svd_bins)
            partial_pdf_filename = pdf_subset.name.filename(
                all_ifos,
                span,
                svd_bin=f"{min_bin}_{max_bin}",
            )
            partial_pdf_path = os.path.join(
                pdf_subset.name.directory(root=rank_dir, start=span[0]),
                partial_pdf_filename,
            )

            partial_pdf_files.append(partial_pdf_path)

            # combine across subset of bins (stage 1)
            round1_layer += Node(
                arguments=Option("marginalize", "ranking-stat-pdf"),
                inputs=Option("input", pdf_subset.files),
                outputs=Option("output", partial_pdf_path),
            )

        executable = "strike-marginalize-likelihood"
        resource_requests = {
            "request_cpus": 1,
            "request_memory": "2GB",
            "request_disk": "1GB",
        }
        round2_layer = create_layer(
            executable,
            condor_config,
            resource_requests,
            name=executable + "-pdf-round-two",
        )

        # combine across all bins (stage 2)
        round2_layer += Node(
            arguments=Option("marginalize", "ranking-stat-pdf"),
            inputs=Option("input", partial_pdf_files),
            outputs=Option("output", marg_pdf_cache.files),
        )
        layers = [round1_layer, round2_layer]
    else:
        round1_layer += Node(
            arguments=Option("marginalize", "ranking-stat-pdf"),
            inputs=Option("input", pdf_cache.files),
            outputs=Option("output", marg_pdf_cache.files),
        )
        layers = [round1_layer]

    return layers


def assign_far(
    condor_config,
    event_config_file,
    trigger_cache,
    marg_pdf_cache,
    post_pdf_cache,
    far_trigger_cache,
):
    executable = "sgnl-extinct-bin"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "6GB",
        "request_disk": "4GB",
    }

    extinct_layer = create_layer(
        executable, condor_config, resource_requests, name=executable + "-round-two"
    )

    executable = "sgnl-assign-far"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "2GB",
        "request_disk": "5GB",
    }
    assign_far_layer = create_layer(executable, condor_config, resource_requests)

    far_triggers = far_trigger_cache.groupby("subtype")

    for subtype, triggers in trigger_cache.groupby("subtype").items():
        if subtype == "":
            extinct_layer += Node(
                inputs=[
                    Option("config-schema", event_config_file),
                    Option("input-database-file", triggers.files),
                    Option("input-rankingstatpdf-file", marg_pdf_cache.files),
                ],
                outputs=[Option("output-rankingstatpdf-file", post_pdf_cache.files)],
            )
        inputs = [
            Option("config-schema", event_config_file),
            Option("input-database-file", triggers.files),
            Option("input-rankingstatpdf-file", post_pdf_cache.files),
        ]
        assign_far_layer += Node(
            arguments=[
                Option("force"),
                Option("tmp-space", condor_scratch_space()),
            ],
            inputs=inputs,
            outputs=[
                Option("output-database-file", far_triggers[subtype].files),
            ],
        )

    return [extinct_layer, assign_far_layer]


def summary_page(
    condor_config,
    event_config_file,
    segments_file,
    segments_name,
    webdir,
    far_trigger_cache,
    seg_far_trigger_cache,
    post_pdf_cache,
    marg_lr_prior_cache,
    mass_model_file,
    injections,
):
    executable = "sgnl-add-segments"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "6GB",
        "request_disk": "4GB",
    }
    seg_layer = create_layer(executable, condor_config, resource_requests)

    if injections:
        executable = "sgnl-sim-page"
        resource_requests = {
            "request_cpus": 1,
            "request_memory": "6GB",
            "request_disk": "10GB",
        }

        sim_layer = create_layer(executable, condor_config, resource_requests)

    executable = "sgnl-results-page"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "6GB",
        "request_disk": "10GB",
    }

    results_layer = create_layer(executable, condor_config, resource_requests)

    far_triggers = far_trigger_cache.groupby("subtype")

    for subtype, seg_triggers in seg_far_trigger_cache.groupby("subtype").items():
        seg_layer += Node(
            arguments=Option("segments-name", segments_name),
            inputs=[
                Option("config-schema", event_config_file),
                Option("input-database-file", far_triggers[subtype].files),
                Option("segments-file", segments_file),
            ],
            outputs=Option("output-database-file", seg_triggers.files),
        )
        if subtype == "":
            results_layer += Node(
                inputs=[
                    Option("config-schema", event_config_file),
                    Option("input-db", seg_triggers.files),
                    Option("input-rank-stat-pdf", post_pdf_cache.files),
                    Option("input-likelihood-file", marg_lr_prior_cache.files),
                    Option(
                        "mass-model-file", mass_model_file, suppress=True, track=False
                    ),
                ],
                outputs=Option("output-html", webdir + "/sgnl-results-page.html"),
            )
        else:
            if injections:
                sim_layer += Node(
                    inputs=[
                        Option("config-schema", event_config_file),
                        Option("input-db", seg_triggers.files),
                    ],
                    outputs=Option(
                        "output-html", webdir + "/sgnl-sim-page-" + subtype + ".html"
                    ),
                )

    return (
        [seg_layer, sim_layer, results_layer]
        if injections
        else [seg_layer, results_layer]
    )


def filter_online(
    psd_config,
    filter_config,
    upload_config,
    services_config,
    source_config,
    condor_config,
    ref_psd_cache,
    svd_bank_cache,
    lr_cache,
    svd_stats,
    zerolag_pdf_cache,
    marg_pdf_cache,
    ifos,
    tag,
    min_instruments,
):
    executable = "sgnl-inspiral"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "5GB",
        "request_disk": "2GB",
    }
    # Set torch-device
    if filter_config.torch_device:
        torch_device = filter_config.torch_device
    else:
        torch_device = "cpu"

    if "cuda" in torch_device:
        resource_requests["request_gpus"] = 1

    layer = create_layer(executable, condor_config, resource_requests, retries=1000)

    assert source_config.data_source in {
        "devshm",
        "arrakis",
    }, "Supported data source options are: devshm, arrakis"

    # set up common options
    common_opts = [
        Option("data-source", source_config.data_source),
        Option("source-queue-timeout", source_config.source_queue_timeout),
        Option("track-psd"),
        Option("psd-fft-length", psd_config.fft_length),
        Option("channel-name", format_ifo_args(ifos, source_config.channel_name)),
        Option(
            "state-channel-name",
            format_ifo_args(ifos, source_config.state_channel_name),
        ),
        Option(
            "state-vector-on-bits",
            format_ifo_args(ifos, source_config.state_vector_on_bits),
        ),
        # Option("tmp-space", dagutil.condor_scratch_space()),
        Option("coincidence-threshold", filter_config.coincidence_threshold),
        Option("analysis-tag", tag),
        Option("far-trials-factor", upload_config.far_trials_factor),
        Option("gracedb-far-threshold", upload_config.gracedb_far_threshold),
        Option(
            "gracedb-aggregator-far-threshold", upload_config.aggregator_far_threshold
        ),
        Option(
            "gracedb-aggregator-far-trials-factor",
            upload_config.aggregator_far_trials_factor,
        ),
        Option("gracedb-group", upload_config.gracedb_group),
        Option("gracedb-search", upload_config.gracedb_search),
        # Option("snr-min", filter_config.snr_min),
    ]

    if filter_config.snapshot_multiprocess:
        common_opts.append(Option("snapshot-multiprocess"))
    if filter_config.snapshot_interval:
        common_opts.append(Option("snapshot-interval", filter_config.snapshot_interval))
    # if filter_config.snapshot_delay:
    #     common_opts.append(Option("snapshot-delay", filter_config.snapshot_delay))

    common_opts.append(Option("min-instruments-candidates", min_instruments))

    # Set shared-memory-dir for devshm
    if source_config.data_source == "devshm":
        common_opts.append(
            Option(
                "shared-memory-dir",
                format_ifo_args(ifos, source_config.shared_memory_dir),
            )
        )

    if filter_config.all_triggers_to_background:
        common_opts.append(Option("all-triggers-to-background"))

    if filter_config.search:
        common_opts.append(Option("search", filter_config.search))

    # Set torch-dtype
    if filter_config.torch_dtype:
        torch_dtype = filter_config.torch_dtype
    else:
        torch_dtype = "float32"
    common_opts.append(Option("torch-dtype", torch_dtype))

    common_opts.append(Option("torch-device", torch_device))

    # Add gstlal CPU upsampling option if enabled
    if filter_config.use_gstlal_cpu_upsample:
        common_opts.append(Option("use-gstlal-cpu-upsample"))

    # Set trigger-finding-duration
    if filter_config.trigger_finding_duration:
        trigger_finding_duration = filter_config.trigger_finding_duration
    else:
        trigger_finding_duration = 1
    common_opts.append(Option("trigger-finding-duration", trigger_finding_duration))

    if services_config.kafka_server:
        common_opts.append(Option("output-kafka-server", services_config.kafka_server))

    if filter_config.cap_singles:
        common_opts.append(Option("cap-singles"))

    if filter_config.verbose:
        common_opts.extend([Option("verbose")])

    # compress ranking stat if requested
    if filter_config.compress_likelihood_ratio:
        common_opts.extend(
            [
                Option("compress-likelihood-ratio"),
                Option(
                    "compress-likelihood-ratio-threshold",
                    filter_config.compress_likelihood_ratio_threshold,
                ),
            ]
        )

    zerolag_pdfs = zerolag_pdf_cache.groupby("bin")

    group_svd_num = None
    if filter_config.group_svd_num:
        group_svd_num = filter_config.group_svd_num
        lr_groups = [lr_group for lr_group in lr_cache.chunked(group_svd_num)]
    elif filter_config.dynamic_group:
        lr_groups = []
        i = 0
        for num in filter_config.dynamic_group.split(","):
            num = int(num)
            lr_groups.append(util.DataCache(lr_cache.name, lr_cache.cache[i : i + num]))
            i = i + num
    else:
        group_svd_num = 1
        lr_groups = [lr_group for lr_group in lr_cache.chunked(group_svd_num)]

    svd_banks = svd_bank_cache.groupby("ifo", "bin")
    for lr_group in lr_groups:
        svd_bins = list(lr_group.groupby("bin").keys())
        if group_svd_num == 1:
            job_tag = f"{int(svd_bins[0]):04d}_noninj"
        else:
            job_tag = f"{int(svd_bins[0]):04d}_{int(svd_bins[-1]):04d}_noninj"

        thresholds = [filter_config.ht_gate_threshold for _ in svd_bins]
        filter_opts = [
            Option("job-tag", job_tag),
            # FIXME: fix gate threshold
            # Option("ht-gate-threshold", calc_gate_threshold(config, svd_bin)),
            # Option("ht-gate-threshold", svd_stats.bins[svd_bin]["ht_gate_threshold"]),
            Option("ht-gate-threshold", thresholds),
        ]
        filter_opts.extend(common_opts)

        svd_bank_files = util.flatten(
            [svd_banks[(ifo, svd_bin)].files for ifo in ifos for svd_bin in svd_bins]
        )

        zerolag_pdf_files = [
            f for svd_bin in svd_bins for f in zerolag_pdfs[(svd_bin)].files
        ]

        inputs = [
            Option("svd-bank", svd_bank_files),
            Option("reference-psd", ref_psd_cache),
            Option("event-config", filter_config.event_config_file),
            Option("input-likelihood-file", lr_group.files),
            Option("rank-stat-pdf", marg_pdf_cache.files),
        ]
        outputs = [
            Option("output-likelihood-file", lr_group.files),
            Option("zerolag-rank-stat-pdf", zerolag_pdf_files),
        ]

        layer += Node(
            arguments=filter_opts,
            inputs=inputs,
            outputs=outputs,
        )

    return layer


def injection_filter_online(
    psd_config,
    filter_config,
    upload_config,
    services_config,
    source_config,
    condor_config,
    ref_psd_cache,
    svd_bank_cache,
    lr_cache,
    svd_stats,
    marg_pdf_cache,
    ifos,
    tag,
    min_instruments,
):
    executable = "sgnl-inspiral"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "5GB",
        "request_disk": "2GB",
    }
    # Set torch-device
    if filter_config.torch_device:
        torch_device = filter_config.torch_device
    else:
        torch_device = "cpu"

    if "cuda" in torch_device:
        resource_requests["request_gpus"] = 1

    layer = create_layer(
        executable,
        condor_config,
        resource_requests,
        retries=1000,
        name=executable + "-inj",
    )

    assert source_config.data_source in {
        "devshm",
        "arrakis",
    }, "Supported data source options are: devshm, arrakis"

    # set up common options
    common_opts = [
        Option("data-source", source_config.data_source),
        Option("source-queue-timeout", source_config.source_queue_timeout),
        Option("track-psd"),
        Option("psd-fft-length", psd_config.fft_length),
        Option("channel-name", format_ifo_args(ifos, source_config.inj_channel_name)),
        Option(
            "state-channel-name",
            format_ifo_args(ifos, source_config.state_channel_name),
        ),
        Option(
            "state-vector-on-bits",
            format_ifo_args(ifos, source_config.state_vector_on_bits),
        ),
        # Option("tmp-space", dagutil.condor_scratch_space()),
        Option("coincidence-threshold", filter_config.coincidence_threshold),
        Option("analysis-tag", tag),
        Option("far-trials-factor", upload_config.far_trials_factor),
        Option("gracedb-far-threshold", upload_config.gracedb_far_threshold),
        Option(
            "gracedb-aggregator-far-threshold", upload_config.aggregator_far_threshold
        ),
        Option(
            "gracedb-aggregator-far-trials-factor",
            upload_config.aggregator_far_trials_factor,
        ),
        Option("gracedb-group", upload_config.gracedb_group),
        Option("gracedb-search", upload_config.inj_gracedb_search),
        Option("injections"),
        # Option("snr-min", filter_config.snr_min),
    ]

    if filter_config.snapshot_multiprocess:
        common_opts.append(Option("snapshot-multiprocess"))
    if filter_config.snapshot_interval:
        common_opts.append(Option("snapshot-interval", filter_config.snapshot_interval))
    # if filter_config.snapshot_delay:
    #     common_opts.append(Option("snapshot-delay", filter_config.snapshot_delay))

    common_opts.append(Option("min-instruments-candidates", min_instruments))

    # Set shared-memory-dir for devshm
    if source_config.data_source == "devshm":
        common_opts.append(
            Option(
                "shared-memory-dir",
                format_ifo_args(ifos, source_config.shared_memory_dir),
            )
        )

    if filter_config.search:
        common_opts.append(Option("search", filter_config.search))

    # Set torch-dtype
    if filter_config.torch_dtype:
        torch_dtype = filter_config.torch_dtype
    else:
        torch_dtype = "float32"
    common_opts.append(Option("torch-dtype", torch_dtype))

    common_opts.append(Option("torch-device", torch_device))

    # Add gstlal CPU upsampling option if enabled
    if filter_config.use_gstlal_cpu_upsample:
        common_opts.append(Option("use-gstlal-cpu-upsample"))

    # Set trigger-finding-duration
    if filter_config.trigger_finding_duration:
        trigger_finding_duration = filter_config.trigger_finding_duration
    else:
        trigger_finding_duration = 1
    common_opts.append(Option("trigger-finding-duration", trigger_finding_duration))

    if services_config.kafka_server:
        common_opts.append(Option("output-kafka-server", services_config.kafka_server))

    if filter_config.cap_singles:
        common_opts.append(Option("cap-singles"))

    if filter_config.verbose:
        common_opts.extend([Option("verbose")])

    # compress ranking stat if requested
    if filter_config.compress_likelihood_ratio:
        common_opts.extend(
            [
                Option("compress-likelihood-ratio"),
                Option(
                    "compress-likelihood-ratio-threshold",
                    filter_config.compress_likelihood_ratio_threshold,
                ),
            ]
        )

    group_svd_num = None
    if filter_config.group_svd_num:
        group_svd_num = filter_config.group_svd_num
        lr_groups = [lr_group for lr_group in lr_cache.chunked(group_svd_num)]
    elif filter_config.dynamic_group:
        lr_groups = []
        i = 0
        for num in filter_config.dynamic_group.split(","):
            num = int(num)
            lr_groups.append(util.DataCache(lr_cache.name, lr_cache.cache[i : i + num]))
            i = i + num
    else:
        group_svd_num = 1
        lr_groups = [lr_group for lr_group in lr_cache.chunked(group_svd_num)]

    svd_banks = svd_bank_cache.groupby("ifo", "bin")
    for lr_group in lr_groups:
        svd_bins = list(lr_group.groupby("bin").keys())
        if group_svd_num == 1:
            job_tag = f"{int(svd_bins[0]):04d}_inj"
        else:
            job_tag = f"{int(svd_bins[0]):04d}_{int(svd_bins[-1]):04d}_inj"

        thresholds = [filter_config.ht_gate_threshold for _ in svd_bins]
        filter_opts = [
            Option("job-tag", job_tag),
            # FIXME: fix gate threshold
            # Option("ht-gate-threshold", calc_gate_threshold(config, svd_bin)),
            # Option("ht-gate-threshold", svd_stats.bins[svd_bin]["ht_gate_threshold"]),
            Option("ht-gate-threshold", thresholds),
        ]
        filter_opts.extend(common_opts)

        svd_bank_files = util.flatten(
            [svd_banks[(ifo, svd_bin)].files for ifo in ifos for svd_bin in svd_bins]
        )

        inputs = [
            Option("svd-bank", svd_bank_files),
            Option("reference-psd", ref_psd_cache),
            Option("event-config", filter_config.event_config_file),
            Option("input-likelihood-file", lr_group.files),
            Option("rank-stat-pdf", marg_pdf_cache.files),
        ]
        layer += Node(
            arguments=filter_opts,
            inputs=inputs,
        )

    return layer


def marginalize_online(
    condor_config,
    filter_config,
    services_config,
    lr_cache,
    tag,
    marg_pdf_cache,
    extinct_percent=None,
    fast_burnin=False,
    calc_pdf_cores=1,
):
    num_cores = calc_pdf_cores if calc_pdf_cores else 1

    executable = "sgnl-ll-marginalize-likelihoods-online"
    resource_requests = {
        "request_cpus": num_cores,
        "request_memory": "4GB",
        "request_disk": "5GB",
    }
    layer = create_layer(executable, condor_config, resource_requests, retries=1000)

    group_svd_num = None
    if filter_config.group_svd_num:
        group_svd_num = filter_config.group_svd_num
        lr_groups = [lr_group for lr_group in lr_cache.chunked(group_svd_num)]
    elif filter_config.dynamic_group:
        lr_groups = []
        i = 0
        for num in filter_config.dynamic_group.split(","):
            num = int(num)
            lr_groups.append(util.DataCache(lr_cache.name, lr_cache.cache[i : i + num]))
            i = i + num
    else:
        group_svd_num = 1
        lr_groups = [lr_group for lr_group in lr_cache.chunked(group_svd_num)]

    svd_bin_groups = [list(lr_group.groupby("bin").keys()) for lr_group in lr_groups]

    if group_svd_num == 1:
        registries = list(
            f"{int(svd_bins[0]):04d}_noninj_registry.txt" for svd_bins in svd_bin_groups
        )
    else:
        registries = list(
            f"{int(svd_bins[0]):04d}_{int(svd_bins[-1]):04d}_noninj_registry.txt"
            for svd_bins in svd_bin_groups
        )
    arguments = [
        Option("registry", registries),
        Option("output", list(marg_pdf_cache.files)),
        Option("output-kafka-server", services_config.kafka_server),
        Option("tag", tag),
        Option("verbose"),
        Option("num-cores", num_cores),
    ]

    if fast_burnin:
        arguments.append(Option("fast-burnin"))

    if extinct_percent:
        arguments.append(Option("extinct-percent", extinct_percent))

    layer += Node(arguments=arguments)

    return layer


def track_noise(
    condor_config,
    source_config,
    filter_config,
    psd_config,
    metrics_config,
    services_config,
    ifos,
    tag,
    ref_psd,
    injection=False,
):
    executable = "sgnl-ll-dq"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "2GB",
        "request_disk": "1GB",
    }
    if injection is True:
        name = executable + "-inj"
    else:
        name = executable

    layer = create_layer(
        executable, condor_config, resource_requests, retries=1000, name=name
    )

    assert source_config.data_source in {
        "devshm",
        "arrakis",
    }, "Supported data source options are: devshm, arrakis"

    for ifo in ifos:
        if injection is True:
            channel_names = [source_config.inj_channel_name[ifo]]
        else:
            channel_names = [source_config.channel_name[ifo]]
        for channel in channel_names:
            # set up datasource options
            arguments = [
                Option("data-source", source_config.data_source),
                Option("source-queue-timeout", source_config.source_queue_timeout),
            ]
            if injection is True:
                arguments.append(Option("injections"))

            arguments.extend(
                [
                    Option("analysis-tag", tag),
                    Option("psd-fft-length", psd_config.fft_length),
                    Option("channel-name", f"{ifo}={channel}"),
                    Option(
                        "state-channel-name",
                        f"{ifo}={source_config.state_channel_name[ifo]}",
                    ),
                    Option(
                        "state-vector-on-bits",
                        f"{ifo}={source_config.state_vector_on_bits[ifo]}",
                    ),
                    Option("reference-psd", ref_psd),
                ]
            )

            # Set shared-memory-dir for devshm
            if source_config.data_source == "devshm":
                if injection is True:
                    memory_location = source_config.inj_shared_memory_dir[ifo]
                else:
                    memory_location = source_config.shared_memory_dir[ifo]
                arguments.extend(
                    [Option("shared-memory-dir", f"{ifo}={memory_location}")]
                )

            if services_config.kafka_server:
                arguments.extend(
                    [Option("output-kafka-server", services_config.kafka_server)]
                )

            layer += Node(arguments=arguments)

    return layer


def count_events(condor_config, services_config, upload_config, tag, zerolag_pdf):
    executable = "sgnl-ll-trigger-counter"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "2GB",
        "request_disk": "1GB",
    }
    layer = create_layer(executable, condor_config, resource_requests, retries=1000)

    layer += Node(
        arguments=[
            Option("kafka-server", services_config.kafka_server),
            Option("output-period", 300),
            Option("topic", "coinc"),
            Option("tag", tag),
        ],
        outputs=Option("output", zerolag_pdf.files),
    )

    return layer


def upload_events(
    condor_config, upload_config, services_config, metrics_config, svd_bins, tag
):
    executable = "sgnl-ll-inspiral-event-uploader"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "2GB",
        "request_disk": "1GB",
    }
    layer = create_layer(executable, condor_config, resource_requests, retries=1000)

    input_topics = (
        ["events", "inj_events"]
        if upload_config.enable_injection_uploads
        else ["events"]
    )

    for input_topic in input_topics:
        arguments = [
            Option("kafka-server", services_config.kafka_server),
            Option("gracedb-group", upload_config.gracedb_group),
            Option("gracedb-pipeline", upload_config.gracedb_pipeline),
            Option("far-threshold", upload_config.aggregator_far_threshold),
            Option("far-trials-factor", upload_config.aggregator_far_trials_factor),
            Option("upload-cadence-type", upload_config.aggregator_cadence_type),
            Option("upload-cadence-factor", upload_config.aggregator_cadence_factor),
            Option("num-jobs", len(svd_bins)),
            Option("tag", tag),
            Option("input-topic", input_topic),
            Option("verbose"),
            Option("scald-config", metrics_config.scald_config),
            Option("max-partitions", 10),
        ]

        # add gracedb service url
        if "inj_" in input_topic:
            arguments.append(
                Option("gracedb-service-url", upload_config.inj_gracedb_service_url)
            )
            arguments.append(Option("gracedb-search", upload_config.inj_gracedb_search))
        else:
            arguments.append(
                Option("gracedb-service-url", upload_config.gracedb_service_url)
            )
            arguments.append(Option("gracedb-search", upload_config.gracedb_search))

        layer += Node(arguments=arguments)

    return layer


def plot_events(condor_config, upload_config, services_config, tag):

    executable = "sgnl-ll-inspiral-event-plotter"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "3GB",
        "request_disk": "1GB",
    }
    layer = create_layer(executable, condor_config, resource_requests, retries=1000)

    upload_topics = (
        ["uploads", "inj_uploads"]
        if upload_config.enable_injection_uploads
        else ["uploads"]
    )
    ranking_stat_topics = (
        ["ranking_stat", "inj_ranking_stat"]
        if upload_config.enable_injection_uploads
        else ["ranking_stat"]
    )

    to_upload = (
        "RANKING_DATA",
        "RANKING_PLOTS",
        "SNR_PLOTS",
        "PSD_PLOTS",
        "DTDPHI_PLOTS",
    )

    for upload_topic, ranking_stat_topic in zip(upload_topics, ranking_stat_topics):
        for upload in to_upload:
            arguments = [
                Option("kafka-server", services_config.kafka_server),
                Option("upload-topic", upload_topic),
                Option("ranking-stat-topic", ranking_stat_topic),
                Option("tag", tag),
                Option("plot", upload),
                Option("verbose"),
            ]

            # add gracedb service url
            if "inj_" in upload_topic:
                arguments.append(
                    Option("gracedb-service-url", upload_config.inj_gracedb_service_url)
                )
            else:
                arguments.append(
                    Option("gracedb-service-url", upload_config.gracedb_service_url)
                )

            layer += Node(arguments=arguments)

    return layer


def collect_metrics(
    condor_config,
    metrics_config,
    services_config,
    filter_config,
    tag,
    ifos,
    svd_bins,
):
    executable = "scald"

    resource_requests = {
        "request_cpus": 1,
        "request_memory": "2GB",
        "request_disk": "1GB",
    }
    metric_leader_layer = create_layer(
        executable,
        condor_config,
        resource_requests,
        retries=1000,
        name="scald_event_collector",
    )

    # FIXME when is this layer used?
    # metric_layer = create_layer(executable, condor_config, resource_requests,
    # retries=1000, name="scald_event_collector")

    # set up common options
    common_opts = [
        Argument("command", "aggregate"),
        Option("config", metrics_config.scald_config),
        Option("uri", f"kafka://{tag}-collect@{services_config.kafka_server}"),
    ]

    # set topic_prefix to distinguish inj and noninj topics
    topic_prefix = ["", "inj_"] if filter_config.injections else [""]

    # define metrics used for aggregation jobs
    snr_metrics = [
        f"{prefix}{ifo}_snr_history" for ifo in ifos for prefix in topic_prefix
    ]
    range_metrics = [f"{prefix}range_history" for prefix in topic_prefix]
    network_metrics = []
    for prefix, topic in list(
        itertools.product(
            topic_prefix,
            ("likelihood_history", "snr_history", "latency_history", "far_history"),
        )
    ):
        network_metrics.append(f"{prefix}{topic}")

    heartbeat_metrics = []
    for prefix, topic in list(
        itertools.product(
            topic_prefix,
            (
                "uptime",
                "event_uploader_heartbeat",
                "event_plotter_heartbeat",
                "pastro_uploader_heartbeat",
            ),
        )
    ):
        heartbeat_metrics.append(f"{prefix}{topic}")
    heartbeat_metrics.append("marginalize_likelihoods_online_heartbeat")
    heartbeat_metrics.append("trigger_counter_heartbeat")

    # state_metrics = [
    #     f"{prefix}{ifo}_strain_dropped" for ifo in ifos for prefix in topic_prefix
    # ]  # FIXME do we need this?
    usage_metrics = [f"{prefix}ram_history" for prefix in topic_prefix]

    latency_metrics = [
        f"{prefix}{ifo}_{stage}_latency"
        for ifo in ifos
        for stage in ("datasource", "whitening", "snrSlice")
        for prefix in topic_prefix
    ]
    latency_metrics.extend([f"{prefix}all_itacacac_latency" for prefix in topic_prefix])

    agg_metrics = list(
        itertools.chain(
            snr_metrics,
            range_metrics,
            network_metrics,
            usage_metrics,
            # state_metrics,
            latency_metrics,
            heartbeat_metrics,
        )
    )

    gates = [
        # f"{gate}segments" for gate in ("statevector", "dqvector", "whiteht", "idq")
        f"{gate}segments"
        for gate in ("statevector", "afterhtgate")
    ]
    seg_metrics = [
        f"{prefix}{ifo}_{gate}"
        for ifo in ifos
        for gate in gates
        for prefix in topic_prefix
    ]

    # set up partitioning
    # FIXME don't hard code the 1000
    max_agg_jobs = 1000

    if filter_config.group_svd_num:
        num_jobs = ceil(len(svd_bins) / filter_config.group_svd_num)
    elif filter_config.dynamic_group:
        num_jobs = len(filter_config.dynamic_group.split(","))
    else:
        num_jobs = len(svd_bins)

    agg_job_bounds = list(range(0, num_jobs, max_agg_jobs))
    min_topics_per_job = 1

    # timeseries metrics
    agg_metrics = groups(
        agg_metrics, max(max_agg_jobs // (4 * num_jobs), min_topics_per_job)
    )
    seg_metrics = groups(
        seg_metrics, max(max_agg_jobs // (4 * num_jobs), min_topics_per_job)
    )

    for metrics in itertools.chain(agg_metrics, seg_metrics):
        for i, _ in enumerate(agg_job_bounds):
            # add jobs to consume each metric
            arguments = list(common_opts)
            topics = []
            schemas = []
            for metric in metrics:
                if "latency_history" in metric:
                    # for latency history we want to
                    # aggregate by max and median so
                    # we need two schemas
                    for aggfunc in ("max", "median"):
                        topics.append(f"sgnl.{tag}.{metric}")
                        schemas.append(f"{metric}_{aggfunc}")
                else:
                    topics.append(f"sgnl.{tag}.{metric}")
                    schemas.append(metric)

            arguments.extend(
                [
                    Option("data-type", "timeseries"),
                    Option("topic", topics),
                    Option("schema", schemas),
                ]
            )

            # elect first metric collector as leader
            if i == 0:
                arguments.append(Option("across-jobs"))
                metric_leader_layer += Node(arguments=arguments)
            else:
                raise ValueError("not implemented")
                # metric_layer += Node(arguments=arguments) # FIXME

    return metric_leader_layer


def collect_metrics_event(
    condor_config,
    metrics_config,
    services_config,
    filter_config,
    tag,
):
    executable = "scald"

    resource_requests = {
        "request_cpus": 1,
        "request_memory": "2GB",
        "request_disk": "1GB",
    }
    event_layer = create_layer(
        executable,
        condor_config,
        resource_requests,
        retries=1000,
        name="scald_event_collector_event",
    )
    # event metrics
    schemas = ["coinc", "inj_coinc"] if filter_config.injections else ["coinc"]

    common_opts = [
        Argument("command", "aggregate"),
        Option("config", metrics_config.scald_config),
        Option("uri", f"kafka://{tag}-collect@{services_config.kafka_server}"),
    ]
    for schema in schemas:
        event_arguments = list(common_opts)
        event_arguments.extend(
            [
                Option("data-type", "triggers"),
                Option("topic", f"sgnl.{tag}.{schema}"),
                Option("schema", schema),
                Option("uri", f"kafka://{tag}-collect@{services_config.kafka_server}"),
            ]
        )
        event_layer += Node(arguments=event_arguments)

    return event_layer


def upload_pastro(
    condor_config, services_config, upload_config, pastro_config, tag, marg_pdf_cache
):

    executable = "sgnl-ll-pastro-uploader"
    resource_requests = {
        "request_cpus": 1,
        "request_memory": "3GB",
        "request_disk": "2GB",
    }

    layer = create_layer(
        executable,
        condor_config,
        resource_requests,
        retries=1000,
    )

    input_topics = (
        ["uploads", "inj_uploads"]
        if upload_config.enable_injection_uploads
        else ["uploads"]
    )

    for model, options in pastro_config.items():
        # make a copy of the model file for injection
        #  jobs to avoid issues with IO
        # FIXME this is so hacky
        if upload_config.enable_injection_uploads:
            path, filename = options.mass_model.split("/")
            inj_mass_model = os.path.join(path, "inj_" + filename)
            shutil.copy(options.mass_model, inj_mass_model)
        for input_topic in input_topics:
            arguments = [
                Option("kafka-server", services_config.kafka_server),
                Option("tag", tag),
                Option("input-topic", input_topic),
                Option("model-name", model),
                Option("pastro-filename", options.upload_file),
                Option(
                    "pastro-model-file",
                    options.mass_model if input_topic == "uploads" else inj_mass_model,
                ),
                Option("rank-stat", marg_pdf_cache.files),
                Option("verbose"),
            ]

            # add gracedb service url
            if "inj_" in input_topic:
                arguments.append(
                    Option("gracedb-service-url", upload_config.inj_gracedb_service_url)
                )
            else:
                arguments.append(
                    Option("gracedb-service-url", upload_config.gracedb_service_url)
                )

            layer += Node(arguments=arguments)

    return layer


def add_likelihood_ratio_file_options(
    svd_bin, svd_stats, prior_config, transfer_only=False
):
    """
    Return a list of options relating to files used for
    terms in the ranking statistic,
    including:
        * dtdphi
        * iDQ timeseries

    if transfer_only is True, do not add options to programs,
    instead, only use these files as inputs for Condor file
    transfer. This is required for jobs that require these files,
    as their paths are tracked through the ranking stat, so
    Condor needs to be aware of these files when not relying on
    a shared file system.
    """
    if transfer_only:
        kwargs = {"track": False, "suppress": True}
    else:
        kwargs = {}

    inputs = [Option("mass-model-file", prior_config.mass_model, **kwargs)]

    if prior_config.idq_timeseries:
        inputs.append(Option("idq-file", prior_config.idq_timeseries, **kwargs))

    if prior_config.dtdphi:
        if isinstance(prior_config.dtdphi, Mapping):
            sub_bank = svd_stats.bins[svd_bin]["bank_name"]
            inputs.append(
                Option("dtdphi-file", prior_config.dtdphi[sub_bank], **kwargs)
            )
        else:
            inputs.append(Option("dtdphi-file", prior_config.dtdphi, **kwargs))

    return inputs


def autocorrelation_length_map(ac_length_range):
    """
    Given autocorrelation length ranges (e.g. 0:15:701)
    or a single autocorrelation value, returns a function that
    maps a given chirp mass to an autocorrelation length.
    """
    if isinstance(ac_length_range, str):
        ac_length_range = [ac_length_range]

    # handle case with AC length ranges
    if isinstance(ac_length_range, Iterable):
        ac_lengths = []
        min_mchirps = []
        max_mchirps = []
        for this_range in ac_length_range:
            min_mchirp, max_mchirp, ac_length = this_range.split(":")
            min_mchirps.append(float(min_mchirp))
            max_mchirps.append(float(max_mchirp))
            ac_lengths.append(int(ac_length))

        # sanity check inputs
        for bound1, bound2 in zip(min_mchirps[1:], max_mchirps[:-1]):
            assert bound1 == bound2, "gaps not allowed in autocorrelation length ranges"

        # convert to binning
        bins = rate.IrregularBins([min_mchirps[0]] + max_mchirps)

    # handle single value case
    else:
        ac_lengths = [ac_length_range]
        bins = rate.IrregularBins([0.0, numpy.inf])

    # create mapping
    def mchirp_to_ac_length(mchirp):
        idx = bins[mchirp]
        return ac_lengths[idx]

    return mchirp_to_ac_length



"""An executable to add additional metadata to svd options file."""

# Copyright (C) 2023  Leo Tsukada (leo.tsukada@ligo.org)
# Copyright (C) 2025  Cody Messick (cody.messick@ligo.org)

import glob
import json
import os
from argparse import ArgumentParser
from collections.abc import Mapping
from typing import Iterable

import numpy
from lal import rate

from sgnl.dags.config import build_config
from sgnl.dags.util import load_svd_options


def parse_command_line():
    parser = ArgumentParser()
    parser.add_argument(
        "-c", "--config", help="Sets the path to read configuration from."
    )
    parser.add_argument(
        "-w",
        "--workflow",
        default="full",
        help="Sets the type of workflow for which to validate config file.",
    )
    parser.add_argument(
        "--dag-dir",
        type=str,
        default=".",
        help="The directory in which to write the dag",
    )

    return parser.parse_args()


def calc_gate_threshold(config, svd_bin, svd_stats, aggregate="max"):
    """
    Given a configuration, svd bin and aggregate, this calculates
    the h(t) gate threshold used for a given svd bin.
    """
    if isinstance(config.filter.ht_gate_threshold, str):
        bank_mchirp = svd_stats["bins"][svd_bin][f"{aggregate}_mchirp"]
        min_mchirp, min_threshold, max_mchirp, max_threshold = [
            float(y)
            for x in config.filter.ht_gate_threshold.split("-")
            for y in x.split(":")
        ]
        gate_mchirp_ratio = (max_threshold - min_threshold) / (max_mchirp - min_mchirp)
        threshold = round(
            gate_mchirp_ratio * (bank_mchirp - min_mchirp) + min_threshold, 3
        )
    else:  # uniform threshold
        threshold = config.filter.ht_gate_threshold
    svd_stats.bins[svd_bin]["ht_gate_threshold"] = threshold
    return threshold


def autocorrelation_length_map(ac_length_range):
    """
    Given autocorrelation length ranges (e.g. 0:15:701)
    or a single autocorrelation value, returns a function that
    maps a given chirp mass to an autocorrelation length.
    """
    if isinstance(ac_length_range, str):
        ac_length_range = [ac_length_range]

    # handle case with AC length ranges
    if isinstance(ac_length_range, Iterable):
        ac_lengths = []
        min_mchirps = []
        max_mchirps = []
        for this_range in ac_length_range:
            min_mchirp, max_mchirp, ac_length = this_range.split(":")
            min_mchirps.append(float(min_mchirp))
            max_mchirps.append(float(max_mchirp))
            ac_lengths.append(int(ac_length))

        # sanity check inputs
        for bound1, bound2 in zip(min_mchirps[1:], max_mchirps[:-1]):
            assert bound1 == bound2, "gaps not allowed in autocorrelation length ranges"

        # convert to binning
        bins = rate.IrregularBins([min_mchirps[0]] + max_mchirps)

    # handle single value case
    else:
        ac_lengths = [ac_length_range]
        bins = rate.IrregularBins([0.0, numpy.inf])

    # create mapping
    def mchirp_to_ac_length(mchirp):
        idx = bins[mchirp]
        return ac_lengths[idx]

    return mchirp_to_ac_length


def svd_bin_to_dtdphi_file(config, svd_bin, stats_bin, aggregate="mean"):
    category_condition = {
        "IMBH": lambda stats_bin: stats_bin[f"{aggregate}_mtotal"] > 100
        and stats_bin[f"{aggregate}_mratio"] < 10,
        "others": lambda stats_bin: True,
    }

    if isinstance(config.prior.dtdphi_file, Mapping):
        if "bank_name" in stats_bin:
            sub_bank = stats_bin["bank_name"]
            dtdphi_file = config.prior.dtdphi_file[sub_bank]
        else:
            assert all(
                category in category_condition
                for category in config.prior.dtdphi_file
                if category != "others"
            ), (
                "At least one of the categories set in config, [%s], is not "
                "defined in those in the source code, [%s]."
                % (
                    ",".join(config.prior.dtdphi_file.keys()),
                    ",".join(category_condition.keys()),
                )
            )
            dtdphi_files = {
                category: filename
                for category, filename in config.prior.dtdphi_file.items()
                if category_condition[category](stats_bin)
            }
            if not len(dtdphi_files):
                raise ValueError(
                    "SVD bin %s does not meet a condition of any category "
                    "given in config.prior.dtdphi_file option. Add 'others' "
                    "category explicitly and point that to the default dtdphi "
                    "file to catch such bins." % (svd_bin,)
                )
            elif "others" in dtdphi_files:
                # pick the default dtdphi file from 'others' category
                dtdphi_file = dtdphi_files.pop("others")
            # With current category_condition (only IMBH + others), at most 1 non-others
            # category can match. This assert guards against future category additions.
            assert len(dtdphi_files) < 2, (
                f"SVD bin {svd_bin} matches multiple categories: "
                f"{','.join(dtdphi_files.keys())}. Add logic to handle this."
            )
            if len(dtdphi_files):
                # pick the dtdphi file specified in the config
                dtdphi_file = [*dtdphi_files.values()][0]
            assert (
                "dtdphi_file" in locals()
            ), "dtdphi_file is not defined even after the validation"
    else:
        dtdphi_file = config.prior.dtdphi_file

    return dtdphi_file


def set_svdbin_option(config):
    _, svd_stats = load_svd_options(config.svd.option_file, config.svd)

    # set up autocorrelation mapping
    mchirp_to_ac_length = autocorrelation_length_map(config.svd.autocorrelation_length)

    for svd_bin, stats_bin in svd_stats.bins.items():
        stats_bin["ac_length"] = mchirp_to_ac_length(stats_bin["mean_mchirp"])
        stats_bin["ht_gate_threshold"] = calc_gate_threshold(config, svd_bin, svd_stats)
        stats_bin["mass_model_file"] = config.prior.mass_model
        if config.prior.idq_timeseries:
            stats_bin["idq_file"] = config.prior.idq_timeseries
        if config.prior.dtdphi_file:
            stats_bin["dtdphi_file"] = svd_bin_to_dtdphi_file(
                config, svd_bin, stats_bin
            )
        if config.filter.reconstruction_segment:
            # FIXME : find a more elegant way to find the files.
            segment_file = glob.glob(
                os.path.join(
                    config.filter.reconstruction_segment,
                    f"*-{svd_bin}_*SEGMENTS-*.xml.gz",
                )
            )
            if len(segment_file) == 0:
                raise ValueError(
                    "No segment file is found in %s"
                    % config.filter.reconstruction_segment
                )
            elif len(segment_file) > 1:
                raise ValueError(
                    "more than one segment file are found in %s"
                    % config.filter.reconstruction_segment
                )
            elif len(segment_file) == 1:
                stats_bin["reconstruction_segment"] = segment_file[0]

        # overwrite options given in seed_likelihood for online rerank
        if config.prior.seed_likelihood:
            for key, option in config.prior.seed_likelihood.items():
                if key != "mass_model":
                    stats_bin[key] = option
                elif config.prior.seed_likelihood.mass_model:
                    stats_bin["mass_model_file"] = (
                        config.prior.seed_likelihood.mass_model
                    )
                    stats_bin["old_mass_model_file"] = config.prior.mass_model

    with open(config.svd.option_file, "w") as jsf:
        jsf.write(json.dumps(svd_stats, sort_keys=True, indent=4))


def main():
    args = parse_command_line()
    config = build_config(args.config, args.dag_dir)
    set_svdbin_option(config)


if __name__ == "__main__":
    main()



"""A file that contains some generic plotting module code"""

# Copyright (C) 2013-2016 Kipp Cannon
# Copyright (C) 2015      Chad Hanna
# Copyright (C) 2023      Cort Posnansky
# Copyright (C) 2024      Yun-Jing Huang

import math
import os
import re

import numpy
from matplotlib.colors import hex2color

golden_ratio = (1.0 + math.sqrt(5.0)) / 2.0


def set_matplotlib_cache_directory():
    """Change the matplotlib cache directory to a local temporary location."""

    # FIXME A cleaner solution would be to set MPLCONFIGDIR in the condor
    # submit file, but doing so is either nontrivial or unsupported.
    if os.getenv("_CONDOR_SCRATCH_DIR"):
        os.environ["MPLCONFIGDIR"] = os.getenv("_CONDOR_SCRATCH_DIR")


def colour_from_instruments(
    instruments: list,
    colours: dict | None = None,
):
    """Get the color for a list of instruments.

    Args:
        instruments:
            list, the instruments
        colours:
            dict, the dictionary of colors as numpy arrays with instruments
            as keys

    Returns:
        array, the color
    """

    if colours is None:
        colours = {
            "G1": numpy.array(hex2color("#222222")),
            "H1": numpy.array(hex2color("#ee0000")),
            "L1": numpy.array(hex2color("#4ba6ff")),
            "V1": numpy.array(hex2color("#9b59b6")),
            "K1": numpy.array(hex2color("#ffb200")),
            "E1": numpy.array((1.0, 0.0, 0.0)),
            "E2": numpy.array((0.0, 0.8, 0.0)),
            "E3": numpy.array((1.0, 0.0, 1.0)),
        }
    # mix colours additively
    colour = sum(map(colours.__getitem__, instruments))
    # use single-instrument colours as-given
    if len(instruments) > 1:
        # desaturate
        colour += len(instruments) - 1
        # normalize
        colour /= colour.max()
    return colour


#
# =============================================================================
#
#                                 TeX Helpers
#
# =============================================================================
#


floatpattern = re.compile("([+-]?[.0-9]+)[Ee]([+-]?[0-9]+)")


def latexnumber(s: str):
    """Convert a string of the form "d.dddde-dd" to "d.dddd \\times 10^{-dd}".  Strings
    that contain neither an "e" nor an "E" are returned unchanged.

    Args:
        s:
            str, the sting to convert

    Example:
        >>> import math
        >>> latexnumber("%.12g" % (math.pi * 1e18))
        '3.14159265359 \\\\times 10^{18}'

    Returns:
        str, the converted string
    """

    if "e" not in s and "E" not in s:
        return s
    match = floatpattern.match(s)
    if match is None:
        return s
    m, e = match.groups()
    return r"$%s \times 10^{%d}$" % (m, int(e))


def latexfilename(s: str):
    """Escapes "\\" and "_" characters, and replaces " " with "~" (non-breaking space).

    Args:
        s:
            str, the string to convert

    Returns:
        str, the converted string
    """

    return s.replace("\\", "\\\\").replace("_", "\\_").replace(" ", "~")



"""An executable to SVD templates."""

# Copyright (C) 2010  Kipp Cannon, Chad Hanna, Leo Singer
# Copyright (C) 2009  Kipp Cannon, Chad Hanna
# Copyright (C) 2015, 2025 Surabhi Sachdev

from argparse import ArgumentParser

import numpy
from igwn_ligolw import lsctables
from igwn_ligolw import utils as ligolw_utils
from lal.utils import CacheEntry

from sgnl import svd_bank
from sgnl.psd import read_psd


def parse_command_line():
    parser = ArgumentParser()
    parser.add_argument(
        "--flow",
        metavar="Hz",
        default=40.0,
        type=float,
        help="Set the template low-frequency cut-off (default = 40.0).",
    )
    parser.add_argument(
        "--sample-rate",
        metavar="Hz",
        type=int,
        help="Set the sample rate. If not set, the sample rate will be based on the "
        "template frequency. The sample rate must be at least twice the highest "
        "frequency in the templates. If provided it must be a power of two",
    )
    parser.add_argument(
        "--padding",
        metavar="pad",
        type=float,
        default=1.5,
        help="Fractional amount to pad time slices.",
    )
    parser.add_argument(
        "--svd-tolerance",
        metavar="match",
        type=float,
        default=0.9999,
        help="Set the SVD reconstruction tolerance (default = 0.9999).",
    )
    parser.add_argument(
        "--reference-psd",
        metavar="filename",
        help="Load the spectrum from this LIGO light-weight XML file (required).",
        required=True,
    )
    parser.add_argument(
        "--instrument-override",
        metavar="ifo",
        help="Override the ifo column of the single inspiral tables",
    )
    parser.add_argument(
        "--template-bank-cache",
        metavar="filename",
        help="Provide a cache file with the names of the LIGO light-weight XML file "
        "from which to load the template bank.",
    )
    parser.add_argument(
        "--write-svd-bank",
        metavar="filename",
        help="Set the filename in which to save the template bank (required).",
        required=True,
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Be verbose (optional)."
    )
    parser.add_argument(
        "--autocorrelation-length",
        type=int,
        default=201,
        help="The minimum number of samples to use for auto-chisquared, default 201 "
        "should be odd",
    )
    parser.add_argument(
        "--samples-min",
        type=int,
        default=1024,
        help="The minimum number of samples to use for time slices default 1024",
    )
    parser.add_argument(
        "--samples-max-256",
        type=int,
        default=1024,
        help="The maximum number of samples to use for time slices with frequencies "
        "above 256Hz, default 1024",
    )
    parser.add_argument(
        "--samples-max-64",
        type=int,
        default=2048,
        help="The maximum number of samples to use for time slices with frequencies "
        "between 64Hz and 256 Hz, default 2048",
    )
    parser.add_argument(
        "--samples-max",
        type=int,
        default=4096,
        help="The maximum number of samples to use for time slices with frequencies "
        "below 64Hz, default 4096",
    )
    parser.add_argument(
        "--max-duration",
        metavar="s",
        type=float,
        default=numpy.inf,
        help="The maximum time duration for the waveform to be searched instead of the "
        "minimum frequency",
    )
    parser.add_argument(
        "--template-banks",  # This captures all remaining arguments
        metavar="Input split banks to perform svd upon",
        nargs="*",  # This allows capturing zero or more template bank files
        help="Provide the split bank files to svd.",
    )

    options = parser.parse_args()

    if not options.autocorrelation_length % 2:
        raise ValueError("--autocorrelation-length must be odd")

    if options.sample_rate is not None and (
        not numpy.log2(options.sample_rate) == int(numpy.log2(options.sample_rate))
    ):
        raise ValueError("--sample-rate must be a power of two")

    return options


def extract_subbank_info(banks, verbose=True):
    var = {"clipleft": [], "clipright": [], "bank-id": []}
    for bank in banks:
        bank_xmldoc = ligolw_utils.load_url(
            bank, contenthandler=svd_bank.DefaultContentHandler, verbose=verbose
        )
        process_params = lsctables.ProcessParamsTable.get_table(bank_xmldoc)
        for row in process_params:
            param = row.param.replace("--", "")
            if param in var:
                var[param].append(row.value)
    assert (
        len(var["clipleft"])
        == len(var["clipright"])
        == len(var["bank-id"])
        == len(banks)
    )
    return var["clipleft"], var["clipright"], var["bank-id"]


def main():
    # FIXME
    snr_min = 0

    # parse arguments
    options = parse_command_line()
    if options.template_bank_cache:
        options.template_banks.extend(
            [CacheEntry(line).url for line in open(options.template_bank_cache)]
        )

    cliplefts, cliprights, bank_ids = extract_subbank_info(options.template_banks)
    cliplefts = [int(cl) for cl in cliplefts]
    cliprights = [int(cr) for cr in cliprights]

    psd = read_psd(options.reference_psd, verbose=options.verbose)

    banks = []
    for template_bank, bank_id, clipleft, clipright in zip(
        options.template_banks, bank_ids, cliplefts, cliprights
    ):
        bank = svd_bank.build_bank(
            template_bank,
            psd,
            options.flow,
            options.max_duration,
            options.svd_tolerance,
            clipleft,
            clipright,
            padding=options.padding,
            verbose=options.verbose,
            snr_threshold=snr_min,
            autocorrelation_length=options.autocorrelation_length,
            samples_min=options.samples_min,
            samples_max_256=options.samples_max_256,
            samples_max_64=options.samples_max_64,
            samples_max=options.samples_max,
            bank_id=bank_id,
            contenthandler=svd_bank.DefaultContentHandler,
            sample_rate=options.sample_rate,
            instrument_override=options.instrument_override,
        )
        banks.append(bank)

    process_param_dict = options.__dict__.copy()
    svd_bank.write_bank(options.write_svd_bank, banks, psd, process_param_dict)


if __name__ == "__main__":
    main()



"""An executable to post a gracedb far upload threshold to inspiral jobs"""

# Copyright (C) 2012  Kipp Cannon
# Copyright (C) 2025  Zach Yarbrough

import argparse

import requests


def parse_command_line():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--set-far-threshold",
        metavar="Hertz",
        type=float,
        help="Set the GraceDB false-alarm rate upload threshold to Hertz",
    )
    parser.add_argument(
        "--analysis-tag",
        type=str,
        required=True,
        help="Unique analysis tag, used in request routes",
    )
    parser.add_argument(
        "--query-far-threshold",
        action="store_true",
        help="Query current far threshold of running jobs",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Be verbose (optional)."
    )
    parser.add_argument(
        "filenames",
        nargs="+",
        help="One or more registry files",
    )

    args = parser.parse_args()

    return args, args.filenames


def get_url(filename):
    return [url.strip() for url in open(filename)][0]


def post_far_threshold(filename, far_threshold, analysis_tag):
    url = get_url(filename)
    post_data = requests.post(
        f"{url}/{analysis_tag}/post/gracedb",
        json={"far-threshold": far_threshold},
        timeout=10,
    )

    return post_data


def query_far_threshold(filename, analysis_tag):
    url = get_url(filename)
    get_data = requests.get(
        f"{url}/{analysis_tag}/get/gracedb/far-threshold",
        timeout=10,
    )

    print(filename, "far threshold:", get_data.json())


def main():
    args, filenames = parse_command_line()

    # argument logic
    if args.query_far_threshold and not args.set_far_threshold:
        for file in filenames:
            query_far_threshold(file, args.analysis_tag)

    elif args.set_far_threshold and not args.query_far_threshold:

        for file in filenames:
            try:
                post_far_threshold(
                    file,
                    args.set_far_threshold,
                    args.analysis_tag,
                )

            except Exception as e:
                print(f"Error when posting far threshold to {file}, error:")
                print(f"{e}")

    else:
        raise ValueError(
            "Improper combination of arguments: specify either"
            "--query-far-threshold or --set-far-threshold, but not both."
        )


if __name__ == "__main__":
    main()



"""Tests for sgnl.transforms.eye_candy module."""

import math
from unittest import mock

from sgnl.transforms import eye_candy


class TestEyeCandyDataclass:
    """Tests for EyeCandy dataclass."""

    def test_eyecandy_creation(self):
        """Test EyeCandy dataclass creation with mocked parent."""
        with mock.patch.object(eye_candy.EyeCandy, "__post_init__"):
            ec = eye_candy.EyeCandy(
                template_sngls=[{"tid1": mock.MagicMock()}],
                event_pad="events",
                state_vector_pads={"H1": "sv_H1", "L1": "sv_L1"},
                ht_gate_pads={"H1": "ht_H1", "L1": "ht_L1"},
            )
            assert ec.event_pad == "events"
            assert "H1" in ec.state_vector_pads


class TestEyeCandyPostInit:
    """Tests for EyeCandy.__post_init__ method."""

    @mock.patch("sgnl.transforms.eye_candy.now", return_value=1000000000000000000)
    @mock.patch.object(eye_candy.TransformElement, "__post_init__")
    def test_post_init_initializes_attributes(self, mock_super_init, mock_now):
        """Test __post_init__ initializes all required attributes."""
        sngl1 = mock.MagicMock()
        sngl2 = mock.MagicMock()
        template_sngls = [{"tid1": sngl1}, {"tid2": sngl2}]

        ec = eye_candy.EyeCandy(
            template_sngls=template_sngls,
            event_pad="events",
            state_vector_pads={"H1": "sv_H1", "L1": "sv_L1"},
            ht_gate_pads={"H1": "ht_H1", "L1": "ht_L1"},
        )

        # Check sink_pad_names
        assert ec.sink_pad_names == ("events", "sv_H1", "sv_L1", "ht_H1", "ht_L1")

        # Check ifos
        assert set(ec.ifos) == {"H1", "L1"}

        # Check sngls_dict
        assert ec.sngls_dict == {"tid1": sngl1, "tid2": sngl2}

        # Check other initializations
        assert ec.outframe is None
        assert ec.startup_time == 1000000000000000000.0
        assert "events" in ec.frames

        mock_super_init.assert_called_once()


class TestEyeCandyPull:
    """Tests for EyeCandy.pull method."""

    @mock.patch("sgnl.transforms.eye_candy.now", return_value=1000000000000000000)
    @mock.patch.object(eye_candy.TransformElement, "__post_init__")
    def test_pull_stores_frame(self, mock_super_init, mock_now):
        """Test pull method stores frames correctly."""
        ec = eye_candy.EyeCandy(
            template_sngls=[{}],
            event_pad="events",
            state_vector_pads={"H1": "sv_H1"},
            ht_gate_pads={"H1": "ht_H1"},
        )
        ec.rsnks = {"pad1": "sv_H1"}

        mock_frame = mock.MagicMock()
        ec.pull("pad1", mock_frame)

        assert ec.frames["sv_H1"] == mock_frame


class TestEyeCandyInternal:
    """Tests for EyeCandy.internal method."""

    def _create_eye_candy(self):
        """Helper to create a properly initialized EyeCandy instance."""
        with mock.patch(
            "sgnl.transforms.eye_candy.now", return_value=2000000000000000000
        ):
            with mock.patch.object(eye_candy.TransformElement, "__post_init__"):
                sngl = mock.MagicMock()
                sngl.mass1 = 1.4
                sngl.mass2 = 1.4
                sngl.spin1z = 0.0
                sngl.spin2z = 0.0

                ec = eye_candy.EyeCandy(
                    template_sngls=[{"tid1": sngl}],
                    event_pad="events",
                    state_vector_pads={"H1": "sv_H1"},
                    ht_gate_pads={"H1": "ht_H1"},
                )
                return ec

    def _create_mock_buffer(self, t0=1000000000000000000, is_gap=False):
        """Helper to create a mock buffer."""
        buf = mock.MagicMock()
        buf.t0 = t0
        buf.is_gap = is_gap
        return buf

    def _create_mock_event_frame(
        self,
        events=None,
        start=1000000000000000000,
        end=1001000000000000000,
        eos=False,
    ):
        """Helper to create a mock event frame."""
        frame = mock.MagicMock()
        frame.events = events if events is not None else []
        frame.start = start
        frame.end = end
        frame.EOS = eos
        return frame

    @mock.patch("sgnl.transforms.eye_candy.now", return_value=2000000000000000000)
    @mock.patch.object(eye_candy.TransformElement, "internal")
    @mock.patch("sgnl.transforms.eye_candy.EventFrame")
    @mock.patch("sgnl.transforms.eye_candy.EventBuffer")
    def test_internal_no_events(
        self, mock_event_buffer, mock_event_frame, mock_super_internal, mock_now
    ):
        """Test internal method when there are no events."""
        ec = self._create_eye_candy()

        # Set up state vector frame
        sv_buf = self._create_mock_buffer(is_gap=False)
        sv_frame = mock.MagicMock()
        sv_frame.__iter__ = mock.MagicMock(return_value=iter([sv_buf]))

        # Set up ht gate frame
        ht_buf = self._create_mock_buffer(is_gap=True)
        ht_frame = mock.MagicMock()
        ht_frame.__iter__ = mock.MagicMock(return_value=iter([ht_buf]))

        # Set up event frame with no events
        event_frame = self._create_mock_event_frame(
            events=[{"max_snr_histories": None, "event": None, "trigger": None}]
        )

        ec.frames = {
            "events": event_frame,
            "sv_H1": sv_frame,
            "ht_H1": ht_frame,
        }

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        ec.internal()

        mock_super_internal.assert_called_once()
        mock_event_frame.assert_called_once()

    @mock.patch("sgnl.transforms.eye_candy.now", return_value=2000000000000000000)
    @mock.patch.object(eye_candy.TransformElement, "internal")
    @mock.patch("sgnl.transforms.eye_candy.EventFrame")
    @mock.patch("sgnl.transforms.eye_candy.EventBuffer")
    def test_internal_with_max_snr_histories(
        self, mock_event_buffer, mock_event_frame, mock_super_internal, mock_now
    ):
        """Test internal with max_snr_histories data."""
        ec = self._create_eye_candy()

        sv_buf = self._create_mock_buffer()
        sv_frame = mock.MagicMock()
        sv_frame.__iter__ = mock.MagicMock(return_value=iter([sv_buf]))

        ht_buf = self._create_mock_buffer()
        ht_frame = mock.MagicMock()
        ht_frame.__iter__ = mock.MagicMock(return_value=iter([ht_buf]))

        # Event with max_snr_histories
        event_frame = self._create_mock_event_frame(
            events=[
                {
                    "max_snr_histories": {"H1": {"time": 1000.0, "snr": 10.5}},
                    "event": None,
                    "trigger": None,
                }
            ]
        )

        ec.frames = {
            "events": event_frame,
            "sv_H1": sv_frame,
            "ht_H1": ht_frame,
        }

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        ec.internal()

        mock_event_frame.assert_called_once()

    @mock.patch("sgnl.transforms.eye_candy.now", return_value=2000000000000000000)
    @mock.patch.object(eye_candy.TransformElement, "internal")
    @mock.patch("sgnl.transforms.eye_candy.EventFrame")
    @mock.patch("sgnl.transforms.eye_candy.EventBuffer")
    def test_internal_with_events_and_triggers(
        self, mock_event_buffer, mock_event_frame, mock_super_internal, mock_now
    ):
        """Test internal with events and triggers."""
        ec = self._create_eye_candy()

        sv_buf = self._create_mock_buffer()
        sv_frame = mock.MagicMock()
        sv_frame.__iter__ = mock.MagicMock(return_value=iter([sv_buf]))

        ht_buf = self._create_mock_buffer()
        ht_frame = mock.MagicMock()
        ht_frame.__iter__ = mock.MagicMock(return_value=iter([ht_buf]))

        # Event with actual events and triggers
        trigger = {
            "_filter_id": "tid1",
            "ifo": "H1",
            "snr": 8.5,
            "chisq": 1.0,
            "phase": 0.5,
        }
        event = {
            "max_snr_histories": None,
            "event": [
                {
                    "network_snr": 12.0,
                    "time": 1000000000000000000,
                    "likelihood": 100.0,
                    "combined_far": 1e-10,
                    "false_alarm_probability": 0.001,
                }
            ],
            "trigger": [[trigger]],
        }
        event_frame = self._create_mock_event_frame(events=[event])

        ec.frames = {
            "events": event_frame,
            "sv_H1": sv_frame,
            "ht_H1": ht_frame,
        }

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        ec.internal()

        mock_event_frame.assert_called_once()

    @mock.patch("sgnl.transforms.eye_candy.now", return_value=2000000000000000000)
    @mock.patch.object(eye_candy.TransformElement, "internal")
    @mock.patch("sgnl.transforms.eye_candy.EventFrame")
    @mock.patch("sgnl.transforms.eye_candy.EventBuffer")
    def test_internal_with_none_likelihood(
        self, mock_event_buffer, mock_event_frame, mock_super_internal, mock_now
    ):
        """Test internal with events that have None likelihood."""
        ec = self._create_eye_candy()

        sv_buf = self._create_mock_buffer()
        sv_frame = mock.MagicMock()
        sv_frame.__iter__ = mock.MagicMock(return_value=iter([sv_buf]))

        ht_buf = self._create_mock_buffer()
        ht_frame = mock.MagicMock()
        ht_frame.__iter__ = mock.MagicMock(return_value=iter([ht_buf]))

        # Event with None likelihood
        event = {
            "max_snr_histories": None,
            "event": [
                {
                    "network_snr": 12.0,
                    "time": 1000000000000000000,
                    "likelihood": None,
                    "combined_far": None,
                    "false_alarm_probability": None,
                }
            ],
            "trigger": [[None]],
        }
        event_frame = self._create_mock_event_frame(events=[event])

        ec.frames = {
            "events": event_frame,
            "sv_H1": sv_frame,
            "ht_H1": ht_frame,
        }

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        ec.internal()

        mock_event_frame.assert_called_once()

    @mock.patch("sgnl.transforms.eye_candy.now", return_value=2000000000000000000)
    @mock.patch.object(eye_candy.TransformElement, "internal")
    @mock.patch("sgnl.transforms.eye_candy.EventFrame")
    @mock.patch("sgnl.transforms.eye_candy.EventBuffer")
    def test_internal_with_inf_likelihood(
        self, mock_event_buffer, mock_event_frame, mock_super_internal, mock_now
    ):
        """Test internal with events that have -inf likelihood."""
        ec = self._create_eye_candy()

        sv_buf = self._create_mock_buffer()
        sv_frame = mock.MagicMock()
        sv_frame.__iter__ = mock.MagicMock(return_value=iter([sv_buf]))

        ht_buf = self._create_mock_buffer()
        ht_frame = mock.MagicMock()
        ht_frame.__iter__ = mock.MagicMock(return_value=iter([ht_buf]))

        # Event with -inf likelihood
        trigger = {
            "_filter_id": "tid1",
            "ifo": "H1",
            "snr": 8.5,
            "chisq": 1.0,
            "phase": 0.5,
        }
        event = {
            "max_snr_histories": None,
            "event": [
                {
                    "network_snr": 12.0,
                    "time": 1000000000000000000,
                    "likelihood": -math.inf,
                    "combined_far": 1e-10,
                    "false_alarm_probability": 0.001,
                }
            ],
            "trigger": [[trigger]],
        }
        event_frame = self._create_mock_event_frame(events=[event])

        ec.frames = {
            "events": event_frame,
            "sv_H1": sv_frame,
            "ht_H1": ht_frame,
        }

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        ec.internal()

        mock_event_frame.assert_called_once()

    @mock.patch("sgnl.transforms.eye_candy.now", return_value=2000000000000000000)
    @mock.patch.object(eye_candy.TransformElement, "internal")
    @mock.patch("sgnl.transforms.eye_candy.EventFrame")
    @mock.patch("sgnl.transforms.eye_candy.EventBuffer")
    def test_internal_multiple_events_multiple_triggers(
        self, mock_event_buffer, mock_event_frame, mock_super_internal, mock_now
    ):
        """Test internal with multiple events and triggers."""
        ec = self._create_eye_candy()

        sv_buf = self._create_mock_buffer()
        sv_frame = mock.MagicMock()
        sv_frame.__iter__ = mock.MagicMock(return_value=iter([sv_buf]))

        ht_buf = self._create_mock_buffer()
        ht_frame = mock.MagicMock()
        ht_frame.__iter__ = mock.MagicMock(return_value=iter([ht_buf]))

        trigger1 = {
            "_filter_id": "tid1",
            "ifo": "H1",
            "snr": 8.5,
            "chisq": 1.0,
            "phase": 0.5,
        }
        trigger2 = {
            "_filter_id": "tid1",
            "ifo": "L1",
            "snr": 7.0,
            "chisq": 1.2,
            "phase": 0.3,
        }

        event1 = {
            "max_snr_histories": {"H1": {"time": 1000.0, "snr": 8.5}},
            "event": [
                {
                    "network_snr": 12.0,
                    "time": 1000000000000000000,
                    "likelihood": 100.0,
                    "combined_far": 1e-10,
                    "false_alarm_probability": 0.001,
                }
            ],
            "trigger": [[trigger1, trigger2]],
        }
        event2 = {
            "max_snr_histories": None,
            "event": [
                {
                    "network_snr": 15.0,
                    "time": 1001000000000000000,
                    "likelihood": 150.0,
                    "combined_far": 1e-12,
                    "false_alarm_probability": 0.0001,
                }
            ],
            "trigger": [[trigger1]],
        }
        event_frame = self._create_mock_event_frame(events=[event1, event2])

        ec.frames = {
            "events": event_frame,
            "sv_H1": sv_frame,
            "ht_H1": ht_frame,
        }

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        ec.internal()

        mock_event_frame.assert_called_once()

    @mock.patch("sgnl.transforms.eye_candy.now", return_value=2000000000000000000)
    @mock.patch.object(eye_candy.TransformElement, "internal")
    @mock.patch("sgnl.transforms.eye_candy.EventFrame")
    @mock.patch("sgnl.transforms.eye_candy.EventBuffer")
    def test_internal_with_none_trigger_in_list(
        self, mock_event_buffer, mock_event_frame, mock_super_internal, mock_now
    ):
        """Test internal with None trigger in trigger list."""
        ec = self._create_eye_candy()

        sv_buf = self._create_mock_buffer()
        sv_frame = mock.MagicMock()
        sv_frame.__iter__ = mock.MagicMock(return_value=iter([sv_buf]))

        ht_buf = self._create_mock_buffer()
        ht_frame = mock.MagicMock()
        ht_frame.__iter__ = mock.MagicMock(return_value=iter([ht_buf]))

        # Mix of None and actual triggers
        trigger1 = {
            "_filter_id": "tid1",
            "ifo": "H1",
            "snr": 8.5,
            "chisq": 1.0,
            "phase": 0.5,
        }
        event = {
            "max_snr_histories": None,
            "event": [
                {
                    "network_snr": 12.0,
                    "time": 1000000000000000000,
                    "likelihood": 100.0,
                    "combined_far": 1e-10,
                    "false_alarm_probability": 0.001,
                }
            ],
            "trigger": [[trigger1, None]],  # Mix of trigger and None
        }
        event_frame = self._create_mock_event_frame(events=[event])

        ec.frames = {
            "events": event_frame,
            "sv_H1": sv_frame,
            "ht_H1": ht_frame,
        }

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        ec.internal()

        mock_event_frame.assert_called_once()

    @mock.patch("sgnl.transforms.eye_candy.now", return_value=2000000000000000000)
    @mock.patch.object(eye_candy.TransformElement, "internal")
    @mock.patch("sgnl.transforms.eye_candy.EventFrame")
    @mock.patch("sgnl.transforms.eye_candy.EventBuffer")
    def test_internal_with_only_far_no_likelihood_history(
        self, mock_event_buffer, mock_event_frame, mock_super_internal, mock_now
    ):
        """Test internal when likelihood is inf but combined_far is valid."""
        ec = self._create_eye_candy()

        sv_buf = self._create_mock_buffer()
        sv_frame = mock.MagicMock()
        sv_frame.__iter__ = mock.MagicMock(return_value=iter([sv_buf]))

        ht_buf = self._create_mock_buffer()
        ht_frame = mock.MagicMock()
        ht_frame.__iter__ = mock.MagicMock(return_value=iter([ht_buf]))

        # Event with positive inf likelihood and valid FAR
        trigger = {
            "_filter_id": "tid1",
            "ifo": "H1",
            "snr": 8.5,
            "chisq": 1.0,
            "phase": 0.5,
        }
        event = {
            "max_snr_histories": None,
            "event": [
                {
                    "network_snr": 12.0,
                    "time": 1000000000000000000,
                    "likelihood": math.inf,
                    "combined_far": 1e-10,
                    "false_alarm_probability": 0.001,
                }
            ],
            "trigger": [[trigger]],
        }
        event_frame = self._create_mock_event_frame(events=[event])

        ec.frames = {
            "events": event_frame,
            "sv_H1": sv_frame,
            "ht_H1": ht_frame,
        }

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        ec.internal()

        mock_event_frame.assert_called_once()


class TestEyeCandyNew:
    """Tests for EyeCandy.new method."""

    @mock.patch("sgnl.transforms.eye_candy.now", return_value=1000000000000000000)
    @mock.patch.object(eye_candy.TransformElement, "__post_init__")
    def test_new_returns_outframe(self, mock_super_init, mock_now):
        """Test new method returns outframe."""
        ec = eye_candy.EyeCandy(
            template_sngls=[{}],
            event_pad="events",
            state_vector_pads={"H1": "sv_H1"},
            ht_gate_pads={"H1": "ht_H1"},
        )
        mock_frame = mock.MagicMock()
        ec.outframe = mock_frame

        result = ec.new("any_pad")

        assert result == mock_frame



import cmath
import math
import sys
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any

import lal
import lalsimulation as lalsim
import numpy
from lal import LIGOTimeGPS
from scipy.linalg import svd as scipy_svd

from sgnl import chirptime, kernels, spawaveform, templates
from sgnl.psd import condition_psd, taperzero_fseries


def tukeywindow(data, samps=200.0):
    assert (
        len(data) >= 2 * samps
    )  # make sure that the user is requesting something sane
    tp = float(samps) / len(data)
    return lal.CreateTukeyREAL8Window(len(data), tp).data.data


def generate_template(
    template_bank_row,
    approximant,
    sample_rate,
    duration,
    f_low,
    f_high,
    amporder=0,
    order=7,
    fwdplan=None,
    fworkspace=None,
):
    """
    Generate a single frequency-domain template, which
    1. is band-limited between f_low and f_high,
    2. has an IFFT which is duration seconds long and
    3. has an IFFT which is sampled at sample_rate Hz
    """
    if approximant not in templates.sgnl_approximants:
        raise ValueError("Unsupported approximant given %s" % approximant)
    assert f_high <= sample_rate // 2

    # FIXME use hcross somday?
    # We don't here because it is not guaranteed to be orthogonal
    # and we add orthogonal phase later

    parameters = {}
    parameters["m1"] = lal.MSUN_SI * template_bank_row.mass1
    parameters["m2"] = lal.MSUN_SI * template_bank_row.mass2
    parameters["S1x"] = template_bank_row.spin1x
    parameters["S1y"] = template_bank_row.spin1y
    parameters["S1z"] = template_bank_row.spin1z
    parameters["S2x"] = template_bank_row.spin2x
    parameters["S2y"] = template_bank_row.spin2y
    parameters["S2z"] = template_bank_row.spin2z
    parameters["distance"] = 1.0e6 * lal.PC_SI
    parameters["inclination"] = 0.0
    parameters["phiRef"] = 0.0
    parameters["longAscNodes"] = 0.0
    parameters["eccentricity"] = 0.0
    parameters["meanPerAno"] = 0.0
    parameters["deltaF"] = 1.0 / duration
    parameters["f_min"] = f_low
    parameters["f_max"] = f_high
    parameters["f_ref"] = 0.0
    parameters["LALparams"] = None
    parameters["approximant"] = lalsim.GetApproximantFromString(str(approximant))

    hplus, hcross = lalsim.SimInspiralFD(**parameters)
    assert len(hplus.data.data) == int(round(f_high * duration)) + 1
    # pad the output vector if the sample rate was higher than the
    # requested final frequency
    if f_high < sample_rate / 2:
        fseries = lal.CreateCOMPLEX16FrequencySeries(
            name=hplus.name,
            epoch=hplus.epoch,
            f0=hplus.f0,
            deltaF=hplus.deltaF,
            length=int(round(sample_rate * duration)) // 2 + 1,
            sampleUnits=hplus.sampleUnits,
        )
        fseries.data.data = numpy.zeros(fseries.data.length)
        fseries.data.data[: hplus.data.length] = hplus.data.data[:]
        hplus = fseries
    return hplus


def condition_imr_template(
    approximant, data, epoch_time, sample_rate_max, max_ringtime
):
    assert (
        -len(data) / sample_rate_max <= epoch_time < 0.0
    ), "Epoch returned follows a different convention"
    # find the index for the peak sample using the epoch returned by
    # the waveform generator
    epoch_index = -int(epoch_time * sample_rate_max) - 1
    # align the peaks according to an overestimate of max rinddown
    # time for a given split bank
    target_index = len(data) - 1 - int(sample_rate_max * max_ringtime)
    # rotate phase so that sample with peak amplitude is real
    phase = numpy.arctan2(data[epoch_index].imag, data[epoch_index].real)
    data *= numpy.exp(-1.0j * phase)
    data = numpy.roll(data, target_index - epoch_index)
    # re-taper the ends of the waveform that got cyclically permuted
    # around the ring
    tukey_beta = 2.0 * abs(target_index - epoch_index) / float(len(data))
    assert 0.0 <= tukey_beta <= 1.0, "waveform got rolled WAY too much"
    data *= lal.CreateTukeyREAL8Window(len(data), tukey_beta).data.data
    # done
    return data, target_index


def condition_ear_warn_template(
    approximant, data, epoch_time, sample_rate_max, max_shift_time
):
    assert (
        -len(data) / sample_rate_max <= epoch_time < 0.0
    ), "Epoch returned follows a different convention"
    # find the index for the peak sample using the epoch returned by
    # the waveform generator
    epoch_index = -int(epoch_time * sample_rate_max) - 1
    # move the early warning waveforms forward according to the waveform
    # that spends the longest in going from fhigh to ISCO in a given
    # split bank. This effectively ends some waveforms at f < fhigh
    target_index = int(sample_rate_max * max_shift_time)
    data = numpy.roll(data, target_index - epoch_index)
    return data, target_index


def compute_autocorrelation_mask(autocorrelation):
    """
    Given an autocorrelation time series, estimate the optimal
    autocorrelation length to use and return a matrix which masks
    out the unwanted elements. FIXME TODO for now just returns
    ones
    """
    return numpy.ones(autocorrelation.shape, dtype="int")


# FIXME FIR_WHITENER variable
@dataclass
class templates_workspace:
    template_table: Sequence[Any]
    approximant: int
    psd: Sequence[Any]
    f_low: float
    time_slices: Sequence[Any]
    autocorrelation_length: int
    fhigh: float
    sample_rate_max: float | None = None
    duration: float | None = None
    length_max: int | None = None
    FIR_WHITENER: int = 0

    def __post_init__(self):
        if self.sample_rate_max is None:
            self.sample_rate_max = max(self.time_slices["rate"])
        if self.fhigh is None:
            self.fhigh = self.sample_rate_max / 2.0
        if self.duration is None:
            self.duration = max(self.time_slices["end"])
        if self.length_max is None:
            self.length_max = int(round(self.duration * self.sample_rate_max))

        # Some input checking to avoid incomprehensible error messages
        if not self.template_table:
            raise ValueError("template list is empty")
        if self.f_low < 0.0:
            raise ValueError("f_low must be >= 0. %s" % repr(self.f_low))

        # working f_low to actually use for generating the waveform.  pick
        # template with lowest chirp mass, compute its duration starting
        # from f_low;  the extra time is 10% of this plus 3 cycles (3 /
        # f_low);  invert to obtain f_low corresponding to desired padding.
        # NOTE:  because SimInspiralChirpStartFrequencyBound() does not
        # account for spin, we set the spins to 0 in the call to
        # SimInspiralChirpTimeBound() regardless of the component's spins.
        template = min(self.template_table, key=lambda row: row.mchirp)
        tchirp = lalsim.SimInspiralChirpTimeBound(
            self.f_low,
            template.mass1 * lal.MSUN_SI,
            template.mass2 * lal.MSUN_SI,
            0.0,
            0.0,
        )
        self.working_f_low = lalsim.SimInspiralChirpStartFrequencyBound(
            1.1 * tchirp + 3.0 / self.f_low,
            template.mass1 * lal.MSUN_SI,
            template.mass2 * lal.MSUN_SI,
        )

        # Add duration of PSD to template length for PSD ringing, round up to power of
        # 2 count of samples
        self.working_length = templates.ceil_pow_2(
            self.length_max + round(1.0 / self.psd.deltaF * self.sample_rate_max)
        )
        self.working_duration = float(self.working_length) / self.sample_rate_max

        if self.psd is not None:
            # Smooth the PSD and interpolate to required resolution
            self.psd = condition_psd(
                self.psd,
                1.0 / self.working_duration,
                minfs=(self.working_f_low, self.f_low),
                maxfs=(self.sample_rate_max / 2.0 * 0.90, self.sample_rate_max / 2.0),
                fir_whiten=self.FIR_WHITENER,
            )

        if self.FIR_WHITENER:
            # Compute a frequency response of the time-domain whitening kernel and
            # effectively taper the psd by zero-ing some elements for a FIR kernel
            self.kernel_fseries = taperzero_fseries(
                kernels.fir_whitener_kernel(
                    self.working_length,
                    self.working_duration,
                    self.sample_rate_max,
                    self.psd,
                ),
                minfs=(self.working_f_low, self.f_low),
                maxfs=(self.sample_rate_max / 2.0 * 0.90, self.sample_rate_max / 2.0),
            )

        self.revplan = lal.CreateReverseCOMPLEX16FFTPlan(self.working_length, 1)
        self.fwdplan = lal.CreateForwardREAL8FFTPlan(self.working_length, 1)
        self.tseries = lal.CreateCOMPLEX16TimeSeries(
            name="timeseries",
            epoch=LIGOTimeGPS(0.0),
            f0=0.0,
            deltaT=1.0 / self.sample_rate_max,
            length=self.working_length,
            sampleUnits=lal.Unit("strain"),
        )
        self.fworkspace = lal.CreateCOMPLEX16FrequencySeries(
            name="template",
            epoch=LIGOTimeGPS(0),
            f0=0.0,
            deltaF=1.0 / self.working_duration,
            length=self.working_length // 2 + 1,
            sampleUnits=lal.Unit("strain s"),
        )

        # Calculate the maximum ring down time or maximum shift time
        if self.approximant in templates.sgnl_IMR_approximants:
            self.max_ringtime = max(
                [
                    chirptime.ringtime(
                        row.mass1 * lal.MSUN_SI + row.mass2 * lal.MSUN_SI,
                        chirptime.overestimate_j_from_chi(max(row.spin1z, row.spin2z)),
                    )
                    for row in self.template_table
                ]
            )
        else:
            if self.sample_rate_max > 2.0 * self.fhigh:
                # Calculate the maximum time we need to shift the early warning
                # waveforms forward by, calculated by the 3.5 approximation from
                # fhigh to ISCO.
                self.max_shift_time = max(
                    [
                        spawaveform.chirptime(
                            row.mass1,
                            row.mass2,
                            7,
                            self.fhigh,
                            0.0,
                            spawaveform.compute_chi(
                                row.mass1, row.mass2, row.spin1z, row.spin2z
                            ),
                        )
                        for row in self.template_table
                    ]
                )

            #
            # Generate each template, downsampling as we go to save memory
            # generate "cosine" component of frequency-domain template.
            # waveform is generated for a canonical distance of 1 Mpc.
            #

    def make_whitened_template(self, template_table_row):
        # FIXME: This is won't work
        # assert template_table_row in self.template_table, "The input
        # Sngl_Inspiral:Table is not found in the workspace."

        # Create template
        fseries = generate_template(
            template_table_row,
            self.approximant,
            self.sample_rate_max,
            self.working_duration,
            self.f_low,
            self.fhigh,
            fwdplan=self.fwdplan,
            fworkspace=self.fworkspace,
        )

        if self.FIR_WHITENER:
            #
            # Compute a product of freq series of the whitening kernel and the template
            # (convolution in time domain) then add quadrature phase
            #
            assert (len(self.kernel_fseries.data.data) // 2 + 1) == len(
                fseries.data.data
            ), "the size of whitening kernel freq series does not match with a given "
            "format of COMPLEX16FrequencySeries."
            fseries.data.data *= self.kernel_fseries.data.data[
                len(self.kernel_fseries.data.data) // 2 - 1 :
            ]
            fseries = templates.QuadraturePhase.add_quadrature_phase(
                fseries, self.working_length
            )
        else:
            #
            # whiten and add quadrature phase ("sine" component)
            #

            if self.psd is not None:
                lal.WhitenCOMPLEX16FrequencySeries(fseries, self.psd)
                fseries = templates.QuadraturePhase.add_quadrature_phase(
                    fseries, self.working_length
                )

        #
        # compute time-domain autocorrelation function
        #

        if self.autocorrelation_length is not None:
            autocorrelation = templates.normalized_autocorrelation(
                fseries, self.revplan
            ).data.data
        else:
            autocorrelation = None

        #
        # transform template to time domain
        #

        lal.COMPLEX16FreqTimeFFT(self.tseries, fseries, self.revplan)

        data = self.tseries.data.data
        epoch_time = fseries.epoch.gpsSeconds + fseries.epoch.gpsNanoSeconds * 1.0e-9

        #
        # extract the portion to be used for filtering
        #

        #
        # condition the template if necessary (e.g. line up IMR
        # waveforms by peak amplitude)
        #

        if self.approximant in templates.sgnl_IMR_approximants:
            data, target_index = condition_imr_template(
                self.approximant,
                data,
                epoch_time,
                self.sample_rate_max,
                self.max_ringtime,
            )
            # record the new end times for the waveforms (since we performed the shifts)
            template_table_row.end = LIGOTimeGPS(
                float(target_index - (len(data) - 1.0)) / self.sample_rate_max
            )
        else:
            if self.sample_rate_max > self.fhigh * 2.0:
                data, target_index = condition_ear_warn_template(
                    self.approximant,
                    data,
                    epoch_time,
                    self.sample_rate_max,
                    self.max_shift_time,
                )
                data *= tukeywindow(data, samps=32)
                # record the new end times for the waveforms (since we performed the
                # shifts)
                template_table_row.end = LIGOTimeGPS(
                    float(target_index) / self.sample_rate_max
                )
            else:
                data *= tukeywindow(data, samps=32)

        data = data[-self.length_max :]

        #
        # normalize so that inner product of template with itself
        # is 2
        #

        norm = abs(numpy.dot(data, numpy.conj(data)))
        data *= cmath.sqrt(2 / norm)

        #
        # sigmasq = 2 \sum_{k=0}^{N-1} |\tilde{h}_{k}|^2 / S_{k} \Delta f
        #
        # XLALWhitenCOMPLEX16FrequencySeries() computed
        #
        # \tilde{h}'_{k} = \sqrt{2 \Delta f} \tilde{h}_{k} / \sqrt{S_{k}}
        #
        # and XLALCOMPLEX16FreqTimeFFT() computed
        #
        # h'_{j} = \Delta f \sum_{k=0}^{N-1} exp(2\pi i j k / N) \tilde{h}'_{k}
        #
        # therefore, "norm" is
        #
        # \sum_{j} |h'_{j}|^{2} = (\Delta f)^{2} \sum_{j} \sum_{k=0}^{N-1}
        # \sum_{k'=0}^{N-1} exp(2\pi i j (k-k') / N) \tilde{h}'_{k} \tilde{h}'^{*}_{k'}
        #                       = (\Delta f)^{2} \sum_{k=0}^{N-1} \sum_{k'=0}^{N-1}
        # \tilde{h}'_{k} \tilde{h}'^{*}_{k'} \sum_{j} exp(2\pi i j (k-k') / N)
        #                       = (\Delta f)^{2} N \sum_{k=0}^{N-1} |\tilde{h}'_{k}|^{2}
        #                       = (\Delta f)^{2} N 2 \Delta f \sum_{k=0}^{N-1}
        # |\tilde{h}_{k}|^{2} / S_{k}
        #                       = (\Delta f)^{2} N sigmasq
        #
        # and \Delta f = 1 / (N \Delta t), so "norm" is
        #
        # \sum_{j} |h'_{j}|^{2} = 1 / (N \Delta t^2) sigmasq
        #
        # therefore
        #
        # sigmasq = norm * N * (\Delta t)^2
        #

        sigmasq = norm * len(data) / self.sample_rate_max**2.0

        return data, autocorrelation, sigmasq


def generate_templates(
    template_table,
    approximant,
    psd,
    f_low,
    time_slices,
    autocorrelation_length=None,
    fhigh=None,
    verbose=False,
):
    # Create workspace for making template bank
    workspace = templates_workspace(
        template_table,
        approximant,
        psd,
        f_low,
        time_slices,
        autocorrelation_length=autocorrelation_length,
        fhigh=fhigh,
    )

    # Check parity of autocorrelation length
    if autocorrelation_length is not None:
        if not (autocorrelation_length % 2):
            raise ValueError(
                f"autocorrelation_length must be odd (got {autocorrelation_length})"
            )
        autocorrelation_bank = numpy.zeros(
            (len(template_table), autocorrelation_length), dtype="cdouble"
        )
        autocorrelation_mask = compute_autocorrelation_mask(autocorrelation_bank)
    else:
        autocorrelation_bank = None
        autocorrelation_mask = None

    # Multiply by 2 * length of the number of sngl_inspiral rows to get the sine/cosine
    # phases.
    template_bank = [
        numpy.zeros(
            (2 * len(template_table), int(round(rate * (end - begin)))), dtype="double"
        )
        for rate, begin, end in time_slices
    ]

    # Store the original normalization of the waveform.  After
    # whitening, the waveforms are normalized.  Use the sigmasq factors
    # to get back the original waveform.
    sigmasq = []

    for i, row in enumerate(template_table):
        if verbose:
            print(
                f"generating template {i + 1}/{len(template_table)}:  "
                f"m1 = {row.mass1:g}, m2 = {row.mass2:g}, "
                f"s1x = {row.spin1x:g}, s1y = {row.spin1y:g}, s1z = {row.spin1z:g}, "
                f"s2x = {row.spin2x:g}, s2y = {row.spin2y:g}, s2z = {row.spin2z:g}",
                file=sys.stderr,
            )
        # FIXME: ensure the row is in template_table
        template, autocorrelation, this_sigmasq = workspace.make_whitened_template(row)

        sigmasq.append(this_sigmasq)

        if autocorrelation is not None:
            autocorrelation_bank[i, ::-1] = numpy.concatenate(
                (
                    autocorrelation[-(autocorrelation_length // 2) :],
                    autocorrelation[: (autocorrelation_length // 2 + 1)],
                )
            )

        #
        # copy real and imaginary parts into adjacent (real-valued)
        # rows of template bank
        #

        for j, time_slice in enumerate(time_slices):
            # start and end times are measured *backwards* from
            # template end;  subtract from n to convert to
            # start and end index;  end:start is the slice to
            # extract, but there's also an amount added equal
            # to 1 less than the stride that I can't explain
            # but probaby has something to do with the reversal
            # of the open/closed boundary conditions through
            # all of this (argh!  Chad!)

            stride = int(round(workspace.sample_rate_max / time_slice["rate"]))
            begin_index = (
                workspace.length_max
                - int(round(time_slice["begin"] * workspace.sample_rate_max))
                + stride
                - 1
            )
            end_index = (
                workspace.length_max
                - int(round(time_slice["end"] * workspace.sample_rate_max))
                + stride
                - 1
            )
            # make sure the rates are commensurate
            assert stride * time_slice["rate"] == workspace.sample_rate_max

            # extract every stride-th sample.  we multiply by
            # \sqrt{stride} to maintain inner product
            # normalization so that the templates still appear
            # to be unit vectors at the reduced sample rate.
            # note that the svd returns unit basis vectors
            # regardless so this factor has no effect on the
            # normalization of the basis vectors used for
            # filtering but it ensures that the chifacs values
            # have the correct relative normalization.
            template_bank[j][(2 * i + 0), :] = template.real[
                end_index:begin_index:stride
            ] * math.sqrt(stride)
            template_bank[j][(2 * i + 1), :] = template.imag[
                end_index:begin_index:stride
            ] * math.sqrt(stride)

    return template_bank, autocorrelation_bank, autocorrelation_mask, sigmasq, workspace


def decompose_templates(template_bank, tolerance, identity=False):
    #
    # sum-of-squares for each template (row).
    #

    chifacs = (template_bank * template_bank).sum(1)

    #
    # this turns this function into a no-op:  the output "basis
    # vectors" are exactly the input templates and the reconstruction
    # matrix is absent (triggers pipeline construction code to omit
    # matrixmixer element)
    #

    if identity:
        return template_bank, None, None, chifacs

    #
    # adjust tolerance according to local norm
    #

    tolerance = 1 - (1 - tolerance) / chifacs.max()

    #
    # S.V.D.
    #

    try:
        U, s, Vh = scipy_svd(template_bank.T)
    except numpy.linalg.LinAlgError as e:
        U, s, Vh = spawaveform.svd(template_bank.T, mod=True, inplace=True)
        print("Falling back on spawaveform", e, file=sys.stderr)

    #
    # determine component count
    #

    residual = numpy.sqrt((s * s).cumsum() / numpy.dot(s, s))
    # FIXME in an ad hoc way force at least 6 principle components
    n = max(min(residual.searchsorted(tolerance) + 1, len(s)), 6)

    #
    # clip decomposition, pre-multiply Vh by s
    #

    U = U[:, :n]
    Vh = numpy.dot(numpy.diag(s), Vh)[:n, :]
    s = s[:n]

    #
    # renormalize the truncated SVD approximation of these template
    # waveform slices making sure their squares still add up to chifacs.
    # This is done by renormalizing the sum of the square of the
    # singular value weighted reconstruction coefficients associated with
    # each template.
    #

    V2 = (Vh * Vh).sum(0)
    for idx, v2 in enumerate(V2):
        Vh[:, idx] *= numpy.sqrt(chifacs[idx] / v2)

    #
    # done.
    #

    return U.T, s, Vh, chifacs



"""Tests for sgnl.bin.flopulator"""

import sys
from unittest import mock

import numpy as np
import pytest


@pytest.fixture(autouse=True)
def mock_dependencies():
    """Mock external dependencies and clean up after tests."""
    # Clear any cached flopulator import first so it will be reimported with mocks
    sys.modules.pop("sgnl.bin.flopulator", None)

    # Store original modules locally within the fixture
    original_modules = {}
    modules_to_mock = ["h5py", "sgnl.svd_bank"]

    # Store original sgnl.svd_bank attribute if sgnl is already imported
    sgnl_mod = sys.modules.get("sgnl")
    original_sgnl_svd_bank = getattr(sgnl_mod, "svd_bank", None) if sgnl_mod else None

    for mod in modules_to_mock:
        original_modules[mod] = sys.modules.get(mod)
        sys.modules[mod] = mock.MagicMock()

    # Also patch the sgnl module's svd_bank attribute if sgnl is imported
    if sgnl_mod is not None:
        sgnl_mod.svd_bank = sys.modules["sgnl.svd_bank"]

    yield

    # Restore originals
    for mod in modules_to_mock:
        if original_modules[mod] is None:
            sys.modules.pop(mod, None)
        else:
            sys.modules[mod] = original_modules[mod]

    # Restore sgnl.svd_bank attribute
    if sgnl_mod is not None and original_sgnl_svd_bank is not None:
        sgnl_mod.svd_bank = original_sgnl_svd_bank

    # Clear the cached import of flopulator so it can be reimported fresh
    sys.modules.pop("sgnl.bin.flopulator", None)


class TestParseCommandLine:
    """Tests for parse_command_line function."""

    def test_required_args(self):
        """Test parsing with required arguments."""
        from sgnl.bin import flopulator

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank.xml",
                "--output-path",
                "output.h5",
            ],
        ):
            options = flopulator.parse_command_line()
            assert options.svd_bank == ["bank.xml"]
            assert options.output_path == "output.h5"

    def test_multiple_svd_banks(self):
        """Test parsing with multiple SVD banks."""
        from sgnl.bin import flopulator

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank1.xml",
                "bank2.xml",
                "bank3.xml",
                "--output-path",
                "output.h5",
            ],
        ):
            options = flopulator.parse_command_line()
            assert options.svd_bank == ["bank1.xml", "bank2.xml", "bank3.xml"]

    def test_memory_flag(self):
        """Test --memory flag."""
        from sgnl.bin import flopulator

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank.xml",
                "--output-path",
                "output.h5",
                "--memory",
            ],
        ):
            options = flopulator.parse_command_line()
            assert options.memory is True

    def test_memory_default(self):
        """Test --memory default is False."""
        from sgnl.bin import flopulator

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank.xml",
                "--output-path",
                "output.h5",
            ],
        ):
            options = flopulator.parse_command_line()
            assert options.memory is False

    def test_verbose_flag(self):
        """Test --verbose flag."""
        from sgnl.bin import flopulator

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank.xml",
                "--output-path",
                "output.h5",
                "--verbose",
            ],
        ):
            options = flopulator.parse_command_line()
            assert options.verbose is True

    def test_verbose_default(self):
        """Test --verbose default is False."""
        from sgnl.bin import flopulator

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank.xml",
                "--output-path",
                "output.h5",
            ],
        ):
            options = flopulator.parse_command_line()
            assert options.verbose is False


class TestFlopulator:
    """Tests for flopulator function."""

    def test_basic_calculation(self):
        """Test basic MFLOPS calculation."""
        from sgnl.bin import flopulator

        r = np.array([256, 512, 1024, 2048])
        rT = np.array([256, 512, 1024, 2048])
        UT = np.array([10, 20, 30, 40])
        MT = 100
        NT = np.array([1000, 500, 250, 125])

        result = flopulator.flopulator(r, rT, UT, MT, NT, verbose=False)

        assert isinstance(result, (int, float, np.floating))
        assert result > 0

    def test_with_verbose(self, capsys):
        """Test flopulator with verbose output."""
        from sgnl.bin import flopulator

        r = np.array([256, 512, 1024, 2048])
        rT = np.array([256, 512, 1024, 2048])
        UT = np.array([10, 20, 30, 40])
        MT = 100
        NT = np.array([1000, 500, 250, 125])

        flopulator.flopulator(r, rT, UT, MT, NT, verbose=True)

        captured = capsys.readouterr()
        assert "MFLOPS from resampling" in captured.out
        assert "MFLOPS from filtering" in captured.out
        assert "MFLOPS from reconstruction" in captured.out
        assert "MFLOPS from addition" in captured.out
        assert "Total MFLOPS" in captured.out
        assert "MFLOPS per complex template" in captured.out
        assert "Ratio number SVD filters" in captured.out

    def test_single_rate(self):
        """Test with single sample rate."""
        from sgnl.bin import flopulator

        r = np.array([1024])
        rT = np.array([1024])
        UT = np.array([50])
        MT = 200
        NT = np.array([500])

        mflops = flopulator.flopulator(r, rT, UT, MT, NT, verbose=False)

        assert isinstance(mflops, (int, float, np.floating))


class TestMemulator:
    """Tests for memulator function."""

    def test_basic_calculation(self):
        """Test basic MB/s calculation."""
        from sgnl.bin import flopulator

        r = np.array([256, 512, 1024, 2048])
        rT = np.array([256, 512, 1024, 2048])
        UT = np.array([10, 20, 30, 40])
        MT = 100
        NT = np.array([1000, 500, 250, 125])

        result = flopulator.memulator(r, rT, UT, MT, NT, verbose=False)

        assert isinstance(result, (int, float, np.floating))
        assert result > 0

    def test_with_verbose(self, capsys):
        """Test memulator with verbose output."""
        from sgnl.bin import flopulator

        r = np.array([256, 512, 1024, 2048])
        rT = np.array([256, 512, 1024, 2048])
        UT = np.array([10, 20, 30, 40])
        MT = 100
        NT = np.array([1000, 500, 250, 125])

        flopulator.memulator(r, rT, UT, MT, NT, verbose=True)

        captured = capsys.readouterr()
        assert "MB/s from resampling" in captured.out
        assert "MB/s from filtering" in captured.out
        assert "MB/s from reconstruction" in captured.out
        assert "MB/s from addition" in captured.out
        assert "Total MB/s" in captured.out
        assert "MB/s per complex template" in captured.out
        assert "Ratio number SVD filters" in captured.out

    def test_single_rate(self):
        """Test with single sample rate."""
        from sgnl.bin import flopulator

        r = np.array([1024])
        rT = np.array([1024])
        UT = np.array([50])
        MT = 200
        NT = np.array([500])

        mbs = flopulator.memulator(r, rT, UT, MT, NT, verbose=False)

        assert isinstance(mbs, (int, float, np.floating))


class TestMain:
    """Tests for main function."""

    def _create_mock_bank_fragment(self, rate, start, end, mix_shape):
        """Create a mock bank fragment."""
        fragment = mock.MagicMock()
        fragment.rate = rate
        fragment.start = start
        fragment.end = end
        fragment.mix_matrix = mock.MagicMock()
        fragment.mix_matrix.shape = mix_shape
        return fragment

    def _create_mock_bank(self, bank_id, fragments):
        """Create a mock bank."""
        bank = mock.MagicMock()
        bank.bank_id = bank_id
        bank.bank_fragments = fragments
        return bank

    def test_main_basic(self, capsys):
        """Test main function with basic options."""
        from sgnl.bin import flopulator

        # Create mock bank fragments
        fragments = [
            self._create_mock_bank_fragment(256, 0.0, 1.0, (10, 100)),
            self._create_mock_bank_fragment(512, 0.0, 0.5, (20, 100)),
        ]
        mock_bank = self._create_mock_bank("0_test", fragments)
        flopulator.svd_bank.read_banks.return_value = [mock_bank]

        # Create mock h5py file
        mock_hf = mock.MagicMock()
        flopulator.h5py.File.return_value = mock_hf

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank.xml",
                "--output-path",
                "output.h5",
            ],
        ):
            flopulator.main()

        flopulator.h5py.File.assert_called_once_with("output.h5", "a")
        flopulator.svd_bank.read_banks.assert_called_once()
        mock_hf.create_group.assert_called()
        mock_hf.close.assert_called_once()

        captured = capsys.readouterr()
        assert "SVD BIN: 0" in captured.out
        assert "Total MFLOPs" in captured.out

    def test_main_with_verbose(self, capsys):
        """Test main function with verbose output."""
        from sgnl.bin import flopulator

        fragments = [
            self._create_mock_bank_fragment(256, 0.0, 1.0, (10, 100)),
            self._create_mock_bank_fragment(512, 0.0, 0.5, (20, 100)),
        ]
        mock_bank = self._create_mock_bank("0_test", fragments)
        flopulator.svd_bank.read_banks.return_value = [mock_bank]

        mock_hf = mock.MagicMock()
        flopulator.h5py.File.return_value = mock_hf

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank.xml",
                "--output-path",
                "output.h5",
                "--verbose",
            ],
        ):
            flopulator.main()

        captured = capsys.readouterr()
        assert "SUB BANK 0" in captured.out
        assert "Unique sampling rates" in captured.out
        assert "Sampling rate for a given time slice" in captured.out
        assert "Total SVD filters" in captured.out
        assert "Number of SVD filter samples" in captured.out
        assert "Total real templates" in captured.out

    def test_main_with_memory(self, capsys):
        """Test main function with memory calculation."""
        from sgnl.bin import flopulator

        fragments = [
            self._create_mock_bank_fragment(256, 0.0, 1.0, (10, 100)),
            self._create_mock_bank_fragment(512, 0.0, 0.5, (20, 100)),
        ]
        mock_bank = self._create_mock_bank("0_test", fragments)
        flopulator.svd_bank.read_banks.return_value = [mock_bank]

        mock_hf = mock.MagicMock()
        flopulator.h5py.File.return_value = mock_hf

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank.xml",
                "--output-path",
                "output.h5",
                "--memory",
            ],
        ):
            flopulator.main()

        captured = capsys.readouterr()
        assert "Total MB/s" in captured.out

    def test_main_multiple_banks(self, capsys):
        """Test main function with multiple SVD banks."""
        from sgnl.bin import flopulator

        fragments1 = [
            self._create_mock_bank_fragment(256, 0.0, 1.0, (10, 100)),
        ]
        mock_bank1 = self._create_mock_bank("0_test", fragments1)

        fragments2 = [
            self._create_mock_bank_fragment(512, 0.0, 0.5, (20, 100)),
        ]
        mock_bank2 = self._create_mock_bank("1_test", fragments2)

        # Reset mock call count and set up side_effect
        flopulator.svd_bank.read_banks.reset_mock()
        flopulator.svd_bank.read_banks.side_effect = [[mock_bank1], [mock_bank2]]

        mock_hf = mock.MagicMock()
        flopulator.h5py.File.return_value = mock_hf

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank1.xml",
                "bank2.xml",
                "--output-path",
                "output.h5",
            ],
        ):
            flopulator.main()

        assert flopulator.svd_bank.read_banks.call_count == 2

    def test_main_multiple_subbanks(self, capsys):
        """Test main function with multiple sub-banks."""
        from sgnl.bin import flopulator

        fragments1 = [
            self._create_mock_bank_fragment(256, 0.0, 1.0, (10, 100)),
        ]
        mock_bank1 = self._create_mock_bank("0_test", fragments1)

        fragments2 = [
            self._create_mock_bank_fragment(512, 0.0, 0.5, (20, 100)),
        ]
        mock_bank2 = self._create_mock_bank("0_test2", fragments2)

        # Return multiple banks for single file (reset side_effect first)
        flopulator.svd_bank.read_banks.side_effect = None
        flopulator.svd_bank.read_banks.return_value = [mock_bank1, mock_bank2]

        mock_hf = mock.MagicMock()
        flopulator.h5py.File.return_value = mock_hf

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank.xml",
                "--output-path",
                "output.h5",
            ],
        ):
            flopulator.main()

        # Verify multiple subbank groups were created
        # Should have: "0" group, "0/subbank-0", "0/subbank-1"
        assert mock_hf.create_group.call_count >= 3

    def test_main_creates_datasets(self):
        """Test that main creates expected datasets."""
        from sgnl.bin import flopulator

        fragments = [
            self._create_mock_bank_fragment(256, 0.0, 1.0, (10, 100)),
        ]
        mock_bank = self._create_mock_bank("0_test", fragments)
        flopulator.svd_bank.read_banks.side_effect = None
        flopulator.svd_bank.read_banks.return_value = [mock_bank]

        mock_hf = mock.MagicMock()
        mock_group = mock.MagicMock()
        mock_hf.create_group.return_value = mock_group
        flopulator.h5py.File.return_value = mock_hf

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank.xml",
                "--output-path",
                "output.h5",
            ],
        ):
            flopulator.main()

        # Check that datasets were created
        dataset_names = [
            call[0][0] for call in mock_group.create_dataset.call_args_list
        ]
        assert "sample-rates" in dataset_names
        assert "num-svd-filters" in dataset_names
        assert "num-real-templates" in dataset_names
        assert "total-mflops" in dataset_names

    def test_main_creates_memory_dataset_when_memory_flag(self):
        """Test that main creates memory dataset when --memory is set."""
        from sgnl.bin import flopulator

        fragments = [
            self._create_mock_bank_fragment(256, 0.0, 1.0, (10, 100)),
        ]
        mock_bank = self._create_mock_bank("0_test", fragments)
        flopulator.svd_bank.read_banks.side_effect = None
        flopulator.svd_bank.read_banks.return_value = [mock_bank]

        mock_hf = mock.MagicMock()
        mock_group = mock.MagicMock()
        mock_hf.create_group.return_value = mock_group
        flopulator.h5py.File.return_value = mock_hf

        with mock.patch(
            "sys.argv",
            [
                "flopulator",
                "--svd-bank",
                "bank.xml",
                "--output-path",
                "output.h5",
                "--memory",
            ],
        ):
            flopulator.main()

        # Check that total-mbs dataset was created
        dataset_names = [
            call[0][0] for call in mock_group.create_dataset.call_args_list
        ]
        assert "total-mbs" in dataset_names



"""Tests for sgnl.viz module."""

import pytest
from matplotlib import pyplot as plt

from sgnl import viz
from sgnl.viz import Section, b64, logo_data, page


class TestB64:
    """Tests for b64 function."""

    def test_b64_with_no_argument(self):
        """Test b64 with default argument (uses current pyplot figure)."""
        # Create a simple plot
        plt.figure()
        plt.plot([1, 2, 3], [1, 2, 3])

        result = b64()

        assert isinstance(result, str)
        # Base64 encoded PNG starts with specific bytes
        assert len(result) > 0
        plt.close()

    def test_b64_with_plot_argument(self):
        """Test b64 with a plot argument."""
        fig, ax = plt.subplots()
        ax.plot([1, 2, 3], [4, 5, 6])

        result = b64(plot=fig)

        assert isinstance(result, str)
        assert len(result) > 0
        plt.close(fig)


class TestPage:
    """Tests for page function."""

    def test_page_with_no_sections(self):
        """Test page with no sections (default argument)."""
        result = page()

        assert isinstance(result, str)
        assert "<!DOCTYPE html>" in result
        assert "SGNL result page" in result
        assert "Streaming Graph Navigator" in result

    def test_page_with_empty_sections_list(self):
        """Test page with an empty sections list."""
        result = page(sections=[])

        assert isinstance(result, str)
        assert "<!DOCTYPE html>" in result

    def test_page_with_sections(self):
        """Test page with actual sections."""
        # Create a section with an image
        section1 = Section(title="Test Section 1", nav="Tab1")
        plt.figure()
        plt.plot([1, 2], [1, 2])
        img_data = b64()
        plt.close()
        section1.append({"img": img_data, "title": "Test Image", "caption": "Caption"})

        # Create a section with a table
        section2 = Section(title="Test Section 2", nav="Tab2")
        section2.append(
            {
                "table": [{"col1": "val1", "col2": "val2"}],
                "title": "Test Table",
                "caption": "Table Caption",
            }
        )

        result = page(sections=[section1, section2])

        assert isinstance(result, str)
        assert "Test Section 1" in result
        assert "Test Section 2" in result
        assert "Tab1" in result
        assert "Tab2" in result
        # First tab should be active
        assert 'class="nav-link active"' in result


class TestSection:
    """Tests for Section class."""

    def test_section_init(self):
        """Test Section initialization."""
        section = Section(title="My Title", nav="My Nav")

        assert section.title == "My Title"
        assert section.nav == "My Nav"
        assert len(section) == 0

    def test_section_html_with_image(self):
        """Test Section.html property with image."""
        section = Section(title="Test", nav="Test Nav")

        plt.figure()
        plt.plot([1, 2], [1, 2])
        img_data = b64()
        plt.close()

        section.append(
            {
                "img": img_data,
                "title": "Image Title",
                "caption": "Image Caption",
            }
        )

        html = section.html

        assert isinstance(html, str)
        assert "Image Title" in html
        assert "Image Caption" in html
        assert "card-img-top" in html

    def test_section_html_with_table(self):
        """Test Section.html property with table."""
        section = Section(title="Test", nav="Test Nav")

        section.append(
            {
                "table": [
                    {"col1": "row1val1", "col2": "row1val2"},
                    {"col1": "row2val1", "col2": "row2val2"},
                ],
                "title": "Table Title",
                "caption": "Table Caption",
            }
        )

        html = section.html

        assert isinstance(html, str)
        assert "Table Title" in html
        assert "Table Caption" in html
        assert "row1val1" in html
        assert "row2val2" in html
        # Column headers default to key names
        assert "col1" in html
        assert "col2" in html

    def test_section_html_with_table_headers(self):
        """Test Section.html property with custom table headers."""
        section = Section(title="Test", nav="Test Nav")

        section.append(
            {
                "table": [{"col1": "val1", "col2": "val2"}],
                "table-headers": {"col1": "Column One", "col2": "Column Two"},
                "title": "Table Title",
                "caption": "Table Caption",
            }
        )

        html = section.html

        assert "Column One" in html
        assert "Column Two" in html

    def test_section_html_invalid_dict_raises(self):
        """Test Section.html raises ValueError for invalid dict."""
        section = Section(title="Test", nav="Test Nav")

        # Dict without 'img' or 'table' key
        section.append(
            {
                "title": "No Image or Table",
                "caption": "This should fail",
            }
        )

        with pytest.raises(ValueError, match="must contain one of"):
            _ = section.html

    def test_section_html_with_empty_table(self):
        """Test Section.html with empty table (table key exists but empty)."""
        section = Section(title="Test", nav="Test Nav")

        section.append(
            {
                "table": [],  # Empty table
                "title": "Empty Table",
                "caption": "This should raise",
            }
        )

        # Empty table should trigger the else branch (ValueError)
        with pytest.raises(ValueError, match="must contain one of"):
            _ = section.html


class TestLogoData:
    """Tests for logo_data function."""

    def test_logo_data_returns_string(self):
        """Test logo_data returns a non-empty string."""
        result = logo_data()

        assert isinstance(result, str)
        assert len(result) > 0

    def test_logo_data_is_base64(self):
        """Test logo_data returns valid base64."""
        import base64

        result = logo_data()

        # Should be valid base64 - this will raise if invalid
        decoded = base64.b64decode(result)
        assert len(decoded) > 0


class TestIfoColors:
    """Tests for IFO color constants."""

    def test_ifo_color_exists(self):
        """Test IFO_COLOR dict has expected keys."""
        assert "H1" in viz.IFO_COLOR
        assert "L1" in viz.IFO_COLOR
        assert "V1" in viz.IFO_COLOR
        assert "K1" in viz.IFO_COLOR

    def test_ifo_combo_color_exists(self):
        """Test IFO_COMBO_COLOR dict is populated."""
        assert len(viz.IFO_COMBO_COLOR) > 0
        # Check some expected combinations
        assert "H1,L1" in viz.IFO_COMBO_COLOR
        assert "H1,L1,V1" in viz.IFO_COMBO_COLOR
